import numpy
import ctypes
cimport numpy

cdef extern from "_discrete.h":
    void _update_pout(int* obs, double* weights, int T, int N, int M, double* pout)

def update_pout(obs, weights, pout, dtype=numpy.float32):
    T = len(obs)
    N, M = numpy.shape(pout)

    # cdef numpy.ndarray[int, ndim=1, mode="c"] obs
    obs = numpy.ascontiguousarray(obs, dtype=ctypes.c_int)

    # pointers to arrays
    if dtype == numpy.float64:
        pobs     = <int*>    numpy.PyArray_DATA(obs)
        pweights = <double*> numpy.PyArray_DATA(weights)
        ppout    = <double*> numpy.PyArray_DATA(pout)
        # call
        _update_pout(pobs, pweights, T, N, M, ppout)
    else:
        raise TypeError




"""Python implementation of Hidden Markov Model kernel functions

This module is considered to be the reference for checking correctness of other
kernels. All implementations are based on paper Rabiners [1].

.. [1] Lawrence R. Rabiner, "A Tutorial on Hidden Markov Models and
   Selected Applications in Speech Recognition", Proceedings of the IEEE,
   vol. 77, issue 2
"""
import numpy as np

from bhmm.hidden import impl_python as ip
from bhmm.hidden import impl_c as ic
from bhmm.util import config


__author__ = "Maikel Nadolski, Christoph Froehner, Frank Noe"
__copyright__ = "Copyright 2015, John D. Chodera and Frank Noe"
__credits__ = ["Maikel Nadolski", "Christoph Froehner", "Frank Noe"]
__license__ = "LGPL"
__maintainer__ = "Frank Noe"
__email__="frank.noe AT fu-berlin DOT de"


# implementation codes
__IMPL_PYTHON__ = 0
__IMPL_C__ = 1

# implementation used
__impl__= __IMPL_PYTHON__


def set_implementation(impl):
    """
    Sets the implementation of this module

    Parameters
    ----------
    impl : str
        One of ["python", "c"]

    """
    global __impl__
    if impl.lower() == 'python':
        __impl__ = __IMPL_PYTHON__
    elif impl.lower() == 'c':
        __impl__ = __IMPL_C__
    else:
        import warnings
        warnings.warn('Implementation '+impl+' is not known. Using the fallback python implementation.')
        __impl__ = __IMPL_PYTHON__



def forward(A, pobs, pi, T=None, alpha_out=None):
    """Compute P( obs | A, B, pi ) and all forward coefficients.

    Parameters
    ----------
    A : ndarray((N,N), dtype = float)
        transition matrix of the hidden states
    pobs : ndarray((T,N), dtype = float)
        pobs[t,i] is the observation probability for observation at time t given hidden state i
    pi : ndarray((N), dtype = float)
        initial distribution of hidden states
    T : int, optional, default = None
        trajectory length. If not given, T = pobs.shape[0] will be used.
    alpha_out : ndarray((T,N), dtype = float), optional, default = None
        containter for the alpha result variables. If None, a new container will be created.

    Returns
    -------
    logprob : float
        The probability to observe the sequence `ob` with the model given
        by `A`, `B` and `pi`.
    alpha : ndarray((T,N), dtype = float), optional, default = None
        alpha[t,i] is the ith forward coefficient of time t. These can be
        used in many different algorithms related to HMMs.

    """
    if __impl__ == __IMPL_PYTHON__:
        return ip.forward(A, pobs, pi, T=T, alpha_out=alpha_out, dtype=config.dtype)
    elif __impl__ == __IMPL_C__:
        return ic.forward(A, pobs, pi, T=T, alpha_out=alpha_out, dtype=config.dtype)
    else:
        raise RuntimeError('Nonexisting implementation selected: '+str(__impl__))


def backward(A, pobs, T=None, beta_out=None):
    """Compute all backward coefficients. With scaling!

    Parameters
    ----------
    A : ndarray((N,N), dtype = float)
        transition matrix of the hidden states
    pobs : ndarray((T,N), dtype = float)
        pobs[t,i] is the observation probability for observation at time t given hidden state i
    T : int, optional, default = None
        trajectory length. If not given, T = pobs.shape[0] will be used.
    beta_out : ndarray((T,N), dtype = float), optional, default = None
        containter for the beta result variables. If None, a new container will be created.

    Returns
    -------
    beta : ndarray((T,N), dtype = float), optional, default = None
        beta[t,i] is the ith backward coefficient of time t. These can be
        used in many different algorithms related to HMMs.

    """
    if __impl__ == __IMPL_PYTHON__:
        return ip.backward(A, pobs, T=T, beta_out=beta_out, dtype=config.dtype)
    elif __impl__ == __IMPL_C__:
        return ic.backward(A, pobs, T=T, beta_out=beta_out, dtype=config.dtype)
    else:
        raise RuntimeError('Nonexisting implementation selected: '+str(__impl__))


def state_probabilities(alpha, beta, T=None, gamma_out=None):
    """ Calculate the (T,N)-probabilty matrix for being in state i at time t.

    Parameters
    ----------
    alpha : ndarray((T,N), dtype = float), optional, default = None
        alpha[t,i] is the ith forward coefficient of time t.
    beta : ndarray((T,N), dtype = float), optional, default = None
        beta[t,i] is the ith forward coefficient of time t.
    T : int, optional, default = None
        trajectory length. If not given, gamma_out.shape[0] will be used. If
        gamma_out is neither given, T = alpha.shape[0] will be used.
    gamma_out : ndarray((T,N), dtype = float), optional, default = None
        containter for the gamma result variables. If None, a new container will be created.

    Returns
    -------
    gamma : ndarray((T,N), dtype = float), optional, default = None
        gamma[t,i] is the probabilty at time t to be in state i !


    See Also
    --------
    forward : to calculate `alpha`
    backward : to calculate `beta`

    """
    if alpha.shape[0] != beta.shape[0]:
        raise ValueError('Inconsistent sizes of alpha and beta.')
    # determine T to use
    if T is None:
        if gamma_out is None:
            T = alpha.shape[0]
        else:
            T = gamma_out.shape[0]
    # compute
    if gamma_out is None:
        gamma_out = alpha * beta
        if T < gamma_out.shape[0]:
            gamma_out = gamma_out[:T]
    else:
        if gamma_out.shape[0] < alpha.shape[0]:
            np.multiply(alpha[:T], beta[:T], gamma_out)
        else:
            np.multiply(alpha, beta, gamma_out)
    # normalize
    np.multiply(gamma_out, 1.0/np.sum(gamma_out, axis=1)[:,None], out = gamma_out)
    # done
    return gamma_out


def state_counts(gamma, T, out = None):
    """ Sum the probabilities of being in state i to time t

    Parameters
    ----------
    gamma : ndarray((T,N), dtype = float), optional, default = None
        gamma[t,i] is the probabilty at time t to be in state i !
    T : int
        number of time steps

    Returns
    -------
    count : numpy.array shape (N)
            count[i] is the summed probabilty to be in state i !

    See Also
    --------
    state_probabilities : to calculate `gamma`

    """
    return np.sum( gamma[0:T], axis = 0, out = out )


# def symbol_counts(gamma, ob, M, dtype=np.float32):
#     """ Sum the observed probabilities to see symbol k in state i.
#
#     Parameters
#     ----------
#     gamma : numpy.array shape (T,N)
#             gamma[t,i] is the probabilty at time t to be in state i !
#     ob : numpy.array shape (T)
#     M : integer. number of possible observationsymbols
#     dtype : item datatype, optional
#
#     Returns
#     -------
#     counts : numpy.array shape (N,M)
#
#     Notes
#     -----
#     This function is independ of alpha and beta being scaled, as long as their
#     scaling is independ in i.
#
#     See Also
#     --------
#     forward, forward_no_scaling : to calculate `alpha`
#     backward, backward_no_scaling : to calculate `beta`
#     """
#     T, N = len(gamma), len(gamma[0])
#     counts = np.zeros((N,M), dtype=type)
#     for t in range(T):
#         for i in range(N):
#             counts[i,ob[t]] += gamma[t,i]
#     return counts


def transition_counts(alpha, beta, A, pobs, T = None, out = None):
    """ Sum for all t the probability to transition from state i to state j.

    Parameters
    ----------
    alpha : ndarray((T,N), dtype = float), optional, default = None
        alpha[t,i] is the ith forward coefficient of time t.
    beta : ndarray((T,N), dtype = float), optional, default = None
        beta[t,i] is the ith forward coefficient of time t.
    A : ndarray((N,N), dtype = float)
        transition matrix of the hidden states
    pobs : ndarray((T,N), dtype = float)
        pobs[t,i] is the observation probability for observation at time t given hidden state i
    T : int
        number of time steps
    out : ndarray((N,N), dtype = float), optional, default = None
        containter for the resulting count matrix. If None, a new matrix will be created.

    Returns
    -------
    counts : numpy.array shape (N, N)
         counts[i, j] is the summed probability to transition from i to j in time [0,T)

    See Also
    --------
    forward : calculate forward coefficients `alpha`
    backward : calculate backward coefficients `beta`

    """
    if __impl__ == __IMPL_PYTHON__:
        return ip.transition_counts(alpha, beta, A, pobs, T=T, out=out, dtype=config.dtype)
    elif __impl__ == __IMPL_C__:
        return ic.transition_counts(alpha, beta, A, pobs, T=T, out=out, dtype=config.dtype)
    else:
        raise RuntimeError('Nonexisting implementation selected: '+str(__impl__))


def viterbi(A, pobs, pi):
    """ Estimate the hidden pathway of maximum likelihood using the Viterbi algorithm.

    Parameters
    ----------
    A : ndarray((N,N), dtype = float)
        transition matrix of the hidden states
    pobs : ndarray((T,N), dtype = float)
        pobs[t,i] is the observation probability for observation at time t given hidden state i
    pi : ndarray((N), dtype = float)
        initial distribution of hidden states

    Returns
    -------
    q : numpy.array shape (T)
        maximum likelihood hidden path

    """
    if __impl__ == __IMPL_PYTHON__:
        return ip.viterbi(A, pobs, pi, dtype=config.dtype)
    elif __impl__ == __IMPL_C__:
        return ic.viterbi(A, pobs, pi, dtype=config.dtype)
    else:
        raise RuntimeError('Nonexisting implementation selected: '+str(__impl__))


def sample_path(alpha, A, pobs, T = None):
    """ Sample the hidden pathway S from the conditional distribution P ( S | Parameters, Observations )

    Parameters
    ----------
    alpha : ndarray((T,N), dtype = float), optional, default = None
        alpha[t,i] is the ith forward coefficient of time t.
    A : ndarray((N,N), dtype = float)
        transition matrix of the hidden states
    pobs : ndarray((T,N), dtype = float)
        pobs[t,i] is the observation probability for observation at time t given hidden state i
    T : int
        number of time steps

    Returns
    -------
    S : numpy.array shape (T)
        maximum likelihood hidden path

    """
    if __impl__ == __IMPL_PYTHON__:
        return ip.sample_path(alpha, A, pobs, T = T, dtype=config.dtype)
    elif __impl__ == __IMPL_C__:
        return ic.sample_path(alpha, A, pobs, T = T, dtype=config.dtype)
    else:
        raise RuntimeError('Nonexisting implementation selected: '+str(__impl__))





"""
Hidden Markov model representation.

"""

import numpy as np
from bhmm.util.logger import logger

__author__ = "John D. Chodera, Frank Noe"
__copyright__ = "Copyright 2015, John D. Chodera and Frank Noe"
__credits__ = ["John D. Chodera", "Frank Noe"]
__license__ = "LGPL"
__maintainer__ = "John D. Chodera"
__email__="jchodera AT gmail DOT com"

from msmtools.estimation import count_matrix

class HMM(object):
    r""" Hidden Markov model (HMM).

    This class is used to represent an HMM. This could be a maximum-likelihood HMM or a sampled HMM from a
    Bayesian posterior.

    Parameters
    ----------
    Tij : np.array with shape (nstates, nstates), optional, default=None
        Row-stochastic transition matrix among states.
    output_model : :class:`bhmm.OutputModel`
        The output model for the states.
    lag : int, optional, default=1
        Lag time (optional) used to estimate the HMM. Used to compute relaxation timescales.
    Pi : np.array with shape (nstates), optional, default=None
        The initial state vector. Required when stationary=False
    stationary : bool, optional, default=True
        If true, the initial distribution is equal to the stationary distribution of the transition matrix
        If false, the initial distribution must be given as Pi
    reversible : bool, optional, default=True
        If true, the transition matrix is reversible.

    Examples
    --------

    >>> # Gaussian HMM
    >>> nstates = 2
    >>> Tij = np.array([[0.8, 0.2], [0.5, 0.5]])
    >>> from bhmm import GaussianOutputModel
    >>> output_model = GaussianOutputModel(nstates, means=[-1, +1], sigmas=[1, 1])
    >>> model = HMM(Tij, output_model)

    >>> # Discrete HMM
    >>> nstates = 2
    >>> Tij = np.array([[0.8, 0.2], [0.5, 0.5]])
    >>> from bhmm import DiscreteOutputModel
    >>> output_model = DiscreteOutputModel([[0.5, 0.1, 0.4], [0.2, 0.3, 0.5]])
    >>> model = HMM(Tij, output_model)

    """
    def __init__(self, Tij, output_model, lag=1, Pi=None, stationary=True, reversible=True):
        # set number of states
        self._nstates = np.array(Tij).shape[0]
        # lag time
        self._lag = lag
        # output model
        self.output_model = output_model
        # hidden state trajectories are optional
        self.hidden_state_trajectories = None
        # parameters
        self._stationary = stationary
        self._reversible = reversible
        # update numbers
        self.update(Tij, Pi)

    def update(self, Tij, Pi=None):
        r""" Updates the transition matrix and recomputes all derived quantities """
        # EMMA imports
        from msmtools import analysis as msmana

        # save a copy of the transition matrix
        self._Tij = np.array(Tij)
        assert msmana.is_transition_matrix(self._Tij), 'Given transition matrix is not a stochastic matrix'
        assert self._Tij.shape[0] == self._nstates, 'Given transition matrix has unexpected number of states '

        # initial / stationary distribution
        if (Pi is not None):
            assert np.all(Pi >= 0), 'Given initial distribution contains negative elements.'
            Pi = np.array(Pi) / np.sum(Pi) # ensure normalization and make a copy

        if (self._stationary):
            pT = msmana.stationary_distribution(self._Tij)
            if Pi is None: # stationary and no stationary distribution fixed, so computing it from trans. mat.
                self._Pi = pT
            else: # stationary but stationary distribution is fixed, so the transition matrix must be consistent
                assert np.allclose(Pi, pT), 'Stationary HMM requested, but given distribution is not the ' \
                                            'stationary distribution of the given transition matrix.'
                self._Pi = Pi
        else:
            if Pi is None: # no initial distribution given, so use stationary distribution anyway
                self._Pi = msmana.stationary_distribution(self._Tij)
            else:
                self._Pi = Pi

        # reversible
        if self._reversible:
            assert msmana.is_reversible(Tij), 'Reversible HMM requested, but given transition matrix is not reversible.'

        # try to do eigendecomposition by default, because it's very cheap for hidden transition matrices
        from scipy.linalg import LinAlgError
        try:
            if self._reversible:
                self._R, self._D, self._L = msmana.rdl_decomposition(self._Tij, norm='reversible')
                # everything must be real-valued
                self._R = self._R.real
                self._D = self._D.real
                self._L = self._L.real
            else:
                self._R, self._D, self._L = msmana.rdl_decomposition(self._Tij, norm='standard')
            self._eigenvalues = np.diag(self._D)
            self._spectral_decomp_available = True
        except LinAlgError:
            logger().warn('Eigendecomposition failed for transition matrix\n'+str(self._Tij)+
                          '\nspectral properties will not be available')
            self._spectral_decomp_available = False

    def __repr__(self):
        from bhmm.output_models import OutputModel
        if issubclass(self.__class__, OutputModel):
            outrepr = repr(OutputModel.__repr__(self))
        else:
            outrepr = repr(self.output_model)
        """ Returns a string representation of the HMM """
        return "HMM(%d, %s, %s, Pi=%s, stationary=%s, reversible=%s)" % (self._nstates,
                                                                         repr(self._Tij),
                                                                         outrepr,
                                                                         repr(self._Pi),
                                                                         repr(self._stationary),
                                                                         repr(self._reversible))

    def __str__(self):
        """ Returns a human-readable string representation of the HMM """
        output  = 'Hidden Markov model\n'
        output += '-------------------\n'
        output += 'nstates: %d\n' % self._nstates
        output += 'Tij:\n'
        output += str(self._Tij) + '\n'
        output += 'Pi:\n'
        output += str(self._Pi) + '\n'
        output += 'output model:\n'
        from bhmm.output_models import OutputModel
        if issubclass(self.__class__, OutputModel):
            output += str(OutputModel.__str__(self))
        else:
            output += str(self.output_model)
        output += '\n'
        return output

    def _assert_spectral_decomposition(self):
        if not self._spectral_decomp_available:
            raise RuntimeError('Trying to access eigenvalues or eigenvectors, but spectral decomposition is not '
                               'available.')

    @property
    def lag(self):
        r""" Lag time of the model, i.e. the number of observated trajectory steps made by the transition matrix """
        return self._lag

    @property
    def is_reversible(self):
        r""" Whether the HMM is reversible """
        return self._reversible

    @property
    def is_stationary(self):
        r""" Whether the MSM is stationary, i.e. whether the initial distribution is the stationary distribution
         of the hidden transition matrix. """
        return self._stationary

    @property
    def nstates(self):
        r""" The number of hidden states """
        return self._nstates

    @property
    def initial_distribution(self):
        r""" The initial distribution of the hidden states """
        return self._Pi

    @property
    def stationary_distribution(self):
        r""" The stationary distribution of hidden states.

        Raises
        ------
        ValueError if the HMM is not stationary

        """
        if self._stationary:
            return self._Pi
        else:
            raise ValueError('HMM is not stationary')

    @property
    def transition_matrix(self):
        r""" The hidden transition matrix """
        return self._Tij

    @property
    def eigenvalues(self):
        r""" Hidden transition matrix eigenvalues

        Returns
        -------
        ts : ndarray(m)
            transition matrix eigenvalues :math:`\lambda_i, i = 1,...,k`., sorted by descending norm.

        """
        self._assert_spectral_decomposition()
        return self._eigenvalues

    @property
    def eigenvectors_left(self):
        r""" Left eigenvectors of the hidden transition matrix

        Returns
        -------
        L : ndarray(nstates,nstates)
            left eigenvectors in a row matrix. l_ij is the j'th component of the i'th left eigenvector

        """
        self._assert_spectral_decomposition()
        return self._L

    @property
    def eigenvectors_right(self):
        r""" Right eigenvectors of the hidden transition matrix

        Returns
        -------
        R : ndarray(nstates,nstates)
            right eigenvectors in a column matrix. r_ij is the i'th component of the j'th right eigenvector

        """
        self._assert_spectral_decomposition()
        return self._R

    @property
    def timescales(self):
        r""" Relaxation timescales of the hidden transition matrix

        Returns
        -------
        ts : ndarray(m)
            relaxation timescales in units of the input trajectory time step,
            defined by :math:`-tau / ln | \lambda_i |, i = 2,...,nstates`, where
            :math:`\lambda_i` are the hidden transition matrix eigenvalues.

        """
        from msmtools.analysis.dense.decomposition import timescales_from_eigenvalues as _timescales

        ts = _timescales(self._eigenvalues, tau=self._lag)
        return ts[1:]

    @property
    def lifetimes(self):
        r""" Lifetimes of states of the hidden transition matrix

        Returns
        -------
        l : ndarray(nstates)
            state lifetimes in units of the input trajectory time step,
            defined by :math:`-tau / ln | p_{ii} |, i = 1,...,nstates`, where
            :math:`p_{ii}` are the diagonal entries of the hidden transition matrix.

        """
        return -self._lag / np.log(np.diag(self.transition_matrix))

    def count_matrix(self, dtype=np.float64):
        # TODO: does this belong here or to the BHMM sampler, or in a subclass containing HMM with data?
        """Compute the transition count matrix from hidden state trajectory.

        Parameters
        ----------
        dtype : numpy.dtype, optional, default=numpy.float64
            The numpy dtype to use for the count matrix.

        Returns
        -------
        C : numpy.array with shape (nstates,nstates)
            C[i,j] is the number of transitions observed from state i to state j

        Raises
        ------
        RuntimeError
            A RuntimeError is raised if the HMM model does not yet have a hidden state trajectory associated with it.

        Examples
        --------

        """

        if self.hidden_state_trajectories is None:
            raise RuntimeError('HMM model does not have a hidden state trajectory.')

        C = count_matrix(self.hidden_state_trajectories, 1, nstates=self._nstates)
        #C = np.zeros((self._nstates,self._nstates), dtype=dtype)
        #for S in self.hidden_state_trajectories:
        #    for t in range(len(S)-1):
        #        C[S[t],S[t+1]] += 1
        return C.toarray()

    # def emission_probability(self, state, observation):
    #     """Compute the emission probability of an observation from a given state.
    #
    #     Parameters
    #     ----------
    #     state : int
    #         The state index for which the emission probability is to be computed.
    #
    #     Returns
    #     -------
    #     Pobs : float
    #         The probability (or probability density, if continuous) of the observation.
    #
    #     TODO
    #     ----
    #     * Vectorize
    #
    #     Examples
    #     --------
    #
    #     Compute the probability of observing an emission of 0 from state 0.
    #
    #     >>> from bhmm import testsystems
    #     >>> model = testsystems.dalton_model(nstates=3)
    #     >>> state_index = 0
    #     >>> observation = 0.0
    #     >>> Pobs = model.emission_probability(state_index, observation)
    #
    #     """
    #     return self.output_model.p_o_i(observation, state)

    # def log_emission_probability(self, state, observation):
    #     """Compute the log emission probability of an observation from a given state.
    #
    #     Parameters
    #     ----------
    #     state : int
    #         The state index for which the emission probability is to be computed.
    #
    #     Returns
    #     -------
    #     log_Pobs : float
    #         The log probability (or probability density, if continuous) of the observation.
    #
    #     TODO
    #     ----
    #     * Vectorize
    #
    #     Examples
    #     --------
    #
    #     Compute the log probability of observing an emission of 0 from state 0.
    #
    #     >>> from bhmm import testsystems
    #     >>> model = testsystems.dalton_model(nstates=3)
    #     >>> state_index = 0
    #     >>> observation = 0.0
    #     >>> log_Pobs = model.log_emission_probability(state_index, observation)
    #
    #     """
    #     return self.output_model.log_p_o_i(observation, state)

    def collect_observations_in_state(self, observations, state_index):
        # TODO: this would work well in a subclass with data
        """Collect a vector of all observations belonging to a specified hidden state.

        Parameters
        ----------
        observations : list of numpy.array
            List of observed trajectories.
        state_index : int
            The index of the hidden state for which corresponding observations are to be retrieved.
        dtype : numpy.dtype, optional, default=numpy.float64
            The numpy dtype to use to store the collected observations.

        Returns
        -------
        collected_observations : numpy.array with shape (nsamples,)
            The collected vector of observations belonging to the specified hidden state.

        Raises
        ------
        RuntimeError
            A RuntimeError is raised if the HMM model does not yet have a hidden state trajectory associated with it.

        """
        if not self.hidden_state_trajectories:
            raise RuntimeError('HMM model does not have a hidden state trajectory.')

        dtype = observations[0].dtype
        collected_observations = np.array([], dtype=dtype)
        for (s_t, o_t) in zip(self.hidden_state_trajectories, observations):
            indices = np.where(s_t == state_index)[0]
            collected_observations = np.append(collected_observations, o_t[indices])

        return collected_observations

    def generate_synthetic_state_trajectory(self, nsteps, initial_Pi=None, start=None, stop=None, dtype=np.int32):
        """Generate a synthetic state trajectory.

        Parameters
        ----------
        nsteps : int
            Number of steps in the synthetic state trajectory to be generated.
        initial_Pi : np.array of shape (nstates,), optional, default=None
            The initial probability distribution, if samples are not to be taken from the intrinsic
            initial distribution.
        start : int
            starting state. Exclusive with initial_Pi
        stop : int
            stopping state. Trajectory will terminate when reaching the stopping state before length number of steps.
        dtype : numpy.dtype, optional, default=numpy.int32
            The numpy dtype to use to store the synthetic trajectory.

        Returns
        -------
        states : np.array of shape (nstates,) of dtype=np.int32
            The trajectory of hidden states, with each element in range(0,nstates).

        Examples
        --------

        Generate a synthetic state trajectory of a specified length.

        >>> from bhmm import testsystems
        >>> model = testsystems.dalton_model()
        >>> states = model.generate_synthetic_state_trajectory(nsteps=100)

        """
        # consistency check
        if initial_Pi is not None and start is not None:
            raise ValueError('Arguments initial_Pi and start are exclusive. Only set one of them.')

        # Generate first state sample.
        if start is None:
            if initial_Pi is not None:
                start = np.random.choice(range(self._nstates), size=1, p=initial_Pi)
            else:
                start = np.random.choice(range(self._nstates), size=1, p=self._Pi)

        # Generate and return trajectory
        from msmtools import generation as msmgen
        traj = msmgen.generate_traj(self.transition_matrix, nsteps, start=start, stop=stop, dt=1)
        return traj.astype(dtype)

    def generate_synthetic_observation(self, state):
        """Generate a synthetic observation from a given state.

        Parameters
        ----------
        state : int
            The index of the state from which the observable is to be sampled.

        Returns
        -------
        observation : float
            The observation from the given state.

        Examples
        --------

        Generate a synthetic observation from a single state.

        >>> from bhmm import testsystems
        >>> model = testsystems.dalton_model()
        >>> observation = model.generate_synthetic_observation(0)

        """
        return self.output_model.generate_observation_from_state(state)

    def generate_synthetic_observation_trajectory(self, length, initial_Pi=None):
        """Generate a synthetic realization of observables.

        Parameters
        ----------
        length : int
            Length of synthetic state trajectory to be generated.
        initial_Pi : np.array of shape (nstates,), optional, default=None
            The initial probability distribution, if samples are not to be taken from equilibrium.

        Returns
        -------
        o_t : np.array of shape (nstates,) of dtype=np.float32
            The trajectory of observations.
        s_t : np.array of shape (nstates,) of dtype=np.int32
            The trajectory of hidden states, with each element in range(0,nstates).

        Examples
        --------

        Generate a synthetic observation trajectory for an equilibrium realization.

        >>> from bhmm import testsystems
        >>> model = testsystems.dalton_model()
        >>> [o_t, s_t] = model.generate_synthetic_observation_trajectory(length=100)

        Use an initial nonequilibrium distribution.

        >>> from bhmm import testsystems
        >>> model = testsystems.dalton_model()
        >>> [o_t, s_t] = model.generate_synthetic_observation_trajectory(length=100, initial_Pi=np.array([1,0,0]))

        """
        # First, generate synthetic state trajetory.
        s_t = self.generate_synthetic_state_trajectory(length, initial_Pi=initial_Pi)

        # Next, generate observations from these states.
        o_t = self.output_model.generate_observation_trajectory(s_t)

        return [o_t, s_t]

    def generate_synthetic_observation_trajectories(self, ntrajectories, length, initial_Pi=None):
        """Generate a number of synthetic realization of observables from this model.

        Parameters
        ----------
        ntrajectories : int
            The number of trajectories to be generated.
        length : int
            Length of synthetic state trajectory to be generated.
        initial_Pi : np.array of shape (nstates,), optional, default=None
            The initial probability distribution, if samples are not to be taken from equilibrium.

        Returns
        -------
        O : list of np.array of shape (nstates,) of dtype=np.float32
            The trajectories of observations
        S : list of np.array of shape (nstates,) of dtype=np.int32
            The trajectories of hidden states

        Examples
        --------

        Generate a number of synthetic trajectories.

        >>> from bhmm import testsystems
        >>> model = testsystems.dalton_model()
        >>> [O, S] = model.generate_synthetic_observation_trajectories(ntrajectories=10, length=100)

        Use an initial nonequilibrium distribution.

        >>> from bhmm import testsystems
        >>> model = testsystems.dalton_model(nstates=3)
        >>> [O, S] = model.generate_synthetic_observation_trajectories(ntrajectories=10, length=100, initial_Pi=np.array([1,0,0]))

        """
        O = list() # observations
        S = list() # state trajectories
        for trajectory_index in range(ntrajectories):
            [o_t, s_t] = self.generate_synthetic_observation_trajectory(length=length, initial_Pi=initial_Pi)
            O.append(o_t)
            S.append(s_t)

        return [O, S]




"""
Sampled Hidden Markov model representation.

"""

__author__ = "John D. Chodera, Frank Noe"
__copyright__ = "Copyright 2015, John D. Chodera and Frank Noe"
__credits__ = ["John D. Chodera", "Frank Noe"]
__license__ = "LGPL"
__maintainer__ = "John D. Chodera"
__email__="jchodera AT gmail DOT com"

import numpy as np

from bhmm.hmm.generic_hmm import HMM
from bhmm.util import config
from bhmm.util.statistics import confidence_interval_arr

class SampledHMM(HMM):
    """ Sampled HMM with a representative single point estimate and error estimates

    Parameters
    ----------
    estimated_hmm : :class:`HMM <generic_hmm.HMM>`
        Representative HMM estimate, e.g. a maximum likelihood estimate or mean HMM.
    sampled_hmms : list of :class:`HMM <generic_hmm.HMM>`
        Sampled HMMs
    conf : float, optional, default = 0.95
        confidence interval, e.g. 0.68 for 1 sigma or 0.95 for 2 sigma.

    """
    def __init__(self, estimated_hmm, sampled_hmms, conf=0.95):
        # call superclass constructer with estimated_hmm
        HMM.__init__(self, estimated_hmm.transition_matrix, estimated_hmm.output_model,
                     lag=estimated_hmm.lag, Pi=estimated_hmm.initial_distribution,
                     stationary=estimated_hmm.is_stationary, reversible=estimated_hmm.is_reversible)
        # save sampled HMMs to calculate statistical moments.
        self._sampled_hmms = sampled_hmms
        self._nsamples = len(sampled_hmms)
        # save confindence interval
        self._conf = conf

    def set_confidence(self, conf):
        r""" Set confidence interval """
        self._conf = conf

    @property
    def nsamples(self):
        r""" Number of samples """
        return self._nsamples

    @property
    def sampled_hmms(self):
        r""" List of sampled HMMs """
        return self._sampled_hmms

    @property
    def confidence_interval(self):
        r""" Confidence interval used """
        return self._conf

    @property
    def initial_distribution_samples(self):
        r""" Samples of the initial distribution """
        res = np.empty((self.nsamples, self.nstates), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i,:] = self._sampled_hmms[i].stationary_distribution
        return res

    @property
    def initial_distribution_mean(self):
        r""" The mean of the initial distribution of the hidden states """
        return np.mean(self.initial_distribution_samples, axis=0)

    @property
    def initial_distribution_std(self):
        r""" The standard deviation of the initial distribution of the hidden states """
        return np.std(self.initial_distribution_samples, axis=0)

    @property
    def initial_distribution_conf(self):
        r""" The standard deviation of the initial distribution of the hidden states """
        return confidence_interval_arr(self.initial_distribution_samples, conf=self._conf)

    @property
    def stationary_distribution_samples(self):
        r""" Samples of the stationary distribution """
        if self._stationary:
            return self.initial_distribution_samples
        else:
            raise ValueError('HMM is not stationary')

    @property
    def stationary_distribution_mean(self):
        r""" The mean of the stationary distribution of the hidden states """
        return np.mean(self.stationary_distribution_samples, axis=0)

    @property
    def stationary_distribution_std(self):
        r""" The standard deviation of the stationary distribution of the hidden states """
        return np.std(self.stationary_distribution_samples, axis=0)

    @property
    def stationary_distribution_conf(self):
        r""" The standard deviation of the stationary distribution of the hidden states """
        return confidence_interval_arr(self.stationary_distribution_samples, conf=self._conf)

    @property
    def transition_matrix_samples(self):
        r""" Samples of the transition matrix """
        res = np.empty((self.nsamples, self.nstates, self.nstates), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i,:,:] = self._sampled_hmms[i].transition_matrix
        return res

    @property
    def transition_matrix_mean(self):
        r""" The mean of the transition_matrix of the hidden states """
        return np.mean(self.transition_matrix_samples, axis=0)

    @property
    def transition_matrix_std(self):
        r""" The standard deviation of the transition_matrix of the hidden states """
        return np.std(self.transition_matrix_samples, axis=0)

    @property
    def transition_matrix_conf(self):
        r""" The standard deviation of the transition_matrix of the hidden states """
        return confidence_interval_arr(self.transition_matrix_samples, conf=self._conf)

    @property
    def eigenvalues_samples(self):
        r""" Samples of the eigenvalues """
        res = np.empty((self.nsamples, self.nstates), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i,:] = self._sampled_hmms[i].eigenvalues
        return res

    @property
    def eigenvalues_mean(self):
        r""" The mean of the eigenvalues of the hidden states """
        return np.mean(self.eigenvalues_samples, axis=0)

    @property
    def eigenvalues_std(self):
        r""" The standard deviation of the eigenvalues of the hidden states """
        return np.std(self.eigenvalues_samples, axis=0)

    @property
    def eigenvalues_conf(self):
        r""" The standard deviation of the eigenvalues of the hidden states """
        return confidence_interval_arr(self.eigenvalues_samples, conf=self._conf)

    @property
    def eigenvectors_left_samples(self):
        r""" Samples of the left eigenvectors of the hidden transition matrix """
        res = np.empty((self.nsamples, self.nstates, self.nstates), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i,:,:] = self._sampled_hmms[i].eigenvectors_left
        return res

    @property
    def eigenvectors_left_mean(self):
        r""" The mean of the left eigenvectors of the hidden transition matrix """
        return np.mean(self.eigenvectors_left_samples, axis=0)

    @property
    def eigenvectors_left_std(self):
        r""" The standard deviation of the left eigenvectors of the hidden transition matrix """
        return np.std(self.eigenvectors_left_samples, axis=0)

    @property
    def eigenvectors_left_conf(self):
        r""" The standard deviation of the left eigenvectors of the hidden transition matrix """
        return confidence_interval_arr(self.eigenvectors_left_samples, conf=self._conf)

    @property
    def eigenvectors_right_samples(self):
        r""" Samples of the right eigenvectors of the hidden transition matrix """
        res = np.empty((self.nsamples, self.nstates, self.nstates), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i,:,:] = self._sampled_hmms[i].eigenvectors_right
        return res

    @property
    def eigenvectors_right_mean(self):
        r""" The mean of the right eigenvectors of the hidden transition matrix """
        return np.mean(self.eigenvectors_right_samples, axis=0)

    @property
    def eigenvectors_right_std(self):
        r""" The standard deviation of the right eigenvectors of the hidden transition matrix """
        return np.std(self.eigenvectors_right_samples, axis=0)

    @property
    def eigenvectors_right_conf(self):
        r""" The standard deviation of the right eigenvectors of the hidden transition matrix """
        return confidence_interval_arr(self.eigenvectors_right_samples, conf=self._conf)

    @property
    def timescales_samples(self):
        r""" Samples of the timescales """
        res = np.empty((self.nsamples, self.nstates-1), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i,:] = self._sampled_hmms[i].timescales
        return res

    @property
    def timescales_mean(self):
        r""" The mean of the timescales of the hidden states """
        return np.mean(self.timescales_samples, axis=0)

    @property
    def timescales_std(self):
        r""" The standard deviation of the timescales of the hidden states """
        return np.std(self.timescales_samples, axis=0)

    @property
    def timescales_conf(self):
        r""" The standard deviation of the timescales of the hidden states """
        return confidence_interval_arr(self.timescales_samples, conf=self._conf)

    @property
    def lifetimes_samples(self):
        r""" Samples of the timescales """
        res = np.empty((self.nsamples, self.nstates), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i,:] = self._sampled_hmms[i].lifetimes
        return res

    @property
    def lifetimes_mean(self):
        r""" The mean of the lifetimes of the hidden states """
        return np.mean(self.lifetimes_samples, axis=0)

    @property
    def lifetimes_std(self):
        r""" The standard deviation of the lifetimes of the hidden states """
        return np.std(self.lifetimes_samples, axis=0)

    @property
    def lifetimes_conf(self):
        r""" The standard deviation of the lifetimes of the hidden states """
        return confidence_interval_arr(self.lifetimes_samples, conf=self._conf)



#!/usr/local/bin/env python

"""
Test provided HMM test system models.

"""

__author__ = "John D. Chodera, Frank Noe"
__copyright__ = "Copyright 2015, John D. Chodera and Frank Noe"
__credits__ = ["John D. Chodera", "Frank Noe"]
__license__ = "FreeBSD"
__maintainer__ = "John D. Chodera"
__email__="jchodera AT gmail DOT com"

import unittest
from bhmm.util import testsystems

class TestTestSystems(unittest.TestCase):

    def test_transition_matrix(self):
        """Test example transition matrices.
        """
        Tij = testsystems.generate_transition_matrix(nstates=3, reversible=False)
        Tij = testsystems.generate_transition_matrix(nstates=3, reversible=True)
        # TODO: Check Tij is proper row-stochastic matrix?
        return

    def test_three_state_model(self):
        """Test three-state model.
        """
        model = testsystems.dalton_model()
        # TODO: Check stationary probiblities are correct?
        return


if __name__=="__main__":
    unittest.main()



r"""Cython implementation of HMM forward backward functions.

.. moduleauthor:: F. Noe <frank DOT noe AT fu-berlin DOT de>

"""
import numpy
import ctypes
cimport numpy

cdef extern from "_hidden.h":
    double _forward(double * alpha, const double *A, const double *pobs, const double *pi, const int N, const int T)

cdef extern from "_hidden.h":
    void _backward(double *beta, const double *A, const double *pobs, const int N, const int T)

# cdef extern from "_hmm.h":
#     void _computeGamma(double *gamma, const double *alpha, const double *beta, const int T, const int N)
#
cdef extern from "_hidden.h":
    void _compute_transition_counts(double *transition_counts, const double *A, const double *pobs, const double *alpha, const double *beta, int N, int T)

cdef extern from "_hidden.h":
    void _compute_viterbi(int *path, const double *A, const double *pobs, const double *pi, int N, int T)

cdef extern from "_hidden.h":
    void _sample_path(int *path, const double *alpha, const double *A, const double *pobs, const int N, const int T)


def cdef_double_array(n1, n2):
    cdef numpy.ndarray[double, ndim=2, mode="c"] out = numpy.zeros( (n1,n2), dtype=numpy.double, order='C' )
    return out


def forward(A, pobs, pi, T=None, alpha_out=None, dtype=numpy.float32):
    # set T
    if (T is None):
        T = pobs.shape[0] # if not set, use the length of pobs as trajectory length
    elif T > pobs.shape[0]:
        raise TypeError('T must be at most the length of pobs.')
    # set N
    N = A.shape[0]
    # prepare alpha array
    # initialize output if necessary
    #cdef numpy.ndarray[double, ndim=2, mode="c"] alpha = numpy.zeros( (T,N), dtype=numpy.double, order='C' )
    if alpha_out is None:
        alpha = cdef_double_array(T,N)
    elif T > alpha_out.shape[0]:
        raise TypeError('alpha_out must at least have length T in order to fit trajectory.')
    else:
        alpha = alpha_out

    #if dtype == numpy.float32:
    #    return ext.forward32(A, pobs, pi)
    if dtype == numpy.float64:
        palpha = <double*> numpy.PyArray_DATA(alpha)
        pA = <double*> numpy.PyArray_DATA(A)
        ppobs = <double*> numpy.PyArray_DATA(pobs)
        ppi = <double*> numpy.PyArray_DATA(pi)
        # call
        logprob = _forward(palpha, pA, ppobs, ppi, N, T)
        return logprob, alpha
    else:
        raise TypeError

def backward(A, pobs, T=None, beta_out=None, dtype=numpy.float32):
    # set T
    if (T is None):
        T = pobs.shape[0] # if not set, use the length of pobs as trajectory length
    elif T > pobs.shape[0]:
        raise ValueError('T must be at most the length of pobs.')
    # set N
    N = A.shape[0]
    #cdef numpy.ndarray[double, ndim=2, mode="c"] beta = numpy.zeros( (T,N), dtype=numpy.double, order='C' )
    # prepare beta array
    if beta_out is None:
        beta = cdef_double_array(T,N)
        #cdef numpy.ndarray[double, ndim=2, mode="c"] beta_out = numpy.zeros( (T,N), dtype=numpy.double, order='C' )
    elif T > beta_out.shape[0]:
        raise ValueError('beta_out must at least have length T in order to fit trajectory.')
    else:
        beta = beta_out

    #if dtype == numpy.float32:
    #    return ext.forward32(A, pobs, pi)
    if dtype == numpy.float64:
        pbeta    = <double*> numpy.PyArray_DATA(beta)
        pA       = <double*> numpy.PyArray_DATA(A)
        ppobs    = <double*> numpy.PyArray_DATA(pobs)
        # call
        _backward(pbeta, pA, ppobs, N, T)
        return beta
    else:
        raise TypeError


# def state_probabilities(alpha, beta, gamma_out=None, dtype=numpy.float32):
#     T = alpha.shape[0]
#     N = alpha.shape[1]
#     # prepare gamma array
#     cdef numpy.ndarray[double, ndim=2, mode="c"] gamma = numpy.zeros( (T,N), dtype=numpy.double, order='C' )
#     #if gamma_out is None:
#     #    gamma_out = cdef_double_array(2,(T,N))
#         #cdef numpy.ndarray[double, ndim=2, mode="c"] gamma_out = numpy.zeros( (T,N), dtype=numpy.double, order='C' )
#
#     #if dtype == numpy.float32:
#     #    return ext.forward32(A, pobs, pi)
#     if dtype == numpy.float64:
#         pgamma   = <double*> numpy.PyArray_DATA(gamma)
#         palpha   = <double*> numpy.PyArray_DATA(alpha)
#         pbeta    = <double*> numpy.PyArray_DATA(beta)
#         # call
#         _computeGamma(pgamma, palpha, pbeta, N, T)
#         return gamma
#     else:
#         raise ValueError
#
#
def transition_counts(alpha, beta, A, pobs, T = None, out = None, dtype=numpy.float32):
    # set T
    if (T is None):
        T = pobs.shape[0] # if not set, use the length of pobs as trajectory length
    elif T > pobs.shape[0]:
        raise ValueError('T must be at most the length of pobs.')
    # set N
    N = len(A)
    # prepare alpha array
    # output
    #cdef numpy.ndarray[double, ndim=2, mode="c"] C = numpy.zeros( (N,N), dtype=numpy.double, order='C' )
    if out is None:
        C = cdef_double_array(N,N)
    else:
        C = out
        #cdef numpy.ndarray[double, ndim=2, mode="c"] out = numpy.zeros( (N,N), dtype=numpy.double, order='C' )

    #if dtype == numpy.float32:
    #    return ext.forward32(A, pobs, pi)
    if dtype == numpy.float64:
        pC     = <double*> numpy.PyArray_DATA(C)
        pA     = <double*> numpy.PyArray_DATA(A)
        ppobs  = <double*> numpy.PyArray_DATA(pobs)
        palpha = <double*> numpy.PyArray_DATA(alpha)
        pbeta  = <double*> numpy.PyArray_DATA(beta)
        # call
        _compute_transition_counts(pC, pA, ppobs, palpha, pbeta, N, T)
        return C
    else:
        raise TypeError


def viterbi(A, pobs, pi, dtype=numpy.float32):
    N = A.shape[0]
    T = pobs.shape[0]
    # prepare path array
    cdef numpy.ndarray[int, ndim=1, mode="c"] path
    path = numpy.zeros( (T), dtype=ctypes.c_int, order='C' )

    #if dtype == numpy.float32:
    #    return ext.forward32(A, pobs, pi)
    if dtype == numpy.float64:
        ppath = <int*>    numpy.PyArray_DATA(path)
        pA    = <double*> numpy.PyArray_DATA(A)
        ppobs = <double*> numpy.PyArray_DATA(pobs)
        ppi   = <double*> numpy.PyArray_DATA(pi)
        # call
        _compute_viterbi(ppath, pA, ppobs, ppi, N, T)
        return path
    else:
        raise TypeError


def sample_path(alpha, A, pobs, T = None, dtype=numpy.float32):
    N = pobs.shape[1]
    # set T
    if (T is None):
        T = pobs.shape[0] # if not set, use the length of pobs as trajectory length
    elif T > pobs.shape[0] or T > alpha.shape[0]:
        raise ValueError('T must be at most the length of pobs and alpha.')
    # prepare path array
    cdef numpy.ndarray[int, ndim=1, mode="c"] path
    path = numpy.zeros( (T), dtype=ctypes.c_int, order='C' )

    if dtype == numpy.float64:
        ppath  = <int*>    numpy.PyArray_DATA(path)
        palpha = <double*> numpy.PyArray_DATA(alpha)
        pA     = <double*> numpy.PyArray_DATA(A)
        ppobs  = <double*> numpy.PyArray_DATA(pobs)
        # call
        _sample_path(ppath, palpha, pA, ppobs, N, T)
        return path
    else:
        raise TypeError



import numpy
import ctypes
cimport numpy

cdef extern from "_gaussian.h":
    void _p_o(const double o, const double* mus, const double* sigmas, const int N, double* p)

cdef extern from "_gaussian.h":
    void _p_obs(const double* o, const double* mus, const double* sigmas, const int N, const int T, double* p)

def cdef_double_vector(n):
    cdef numpy.ndarray[double, ndim=1, mode="c"] out = numpy.zeros( (n), dtype=ctypes.c_double, order='C' )
    return out

def cdef_single_vector(n):
    cdef numpy.ndarray[float, ndim=1, mode="c"] out = numpy.zeros( (n), dtype=ctypes.c_float, order='C' )
    return out

def cdef_double_matrix(n1,n2):
    cdef numpy.ndarray[double, ndim=2, mode="c"] out = numpy.zeros( (n1,n2), dtype=ctypes.c_double, order='C' )
    return out

def cdef_single_matrix(n1,n2):
    cdef numpy.ndarray[float, ndim=2, mode="c"] out = numpy.zeros( (n1,n2), dtype=ctypes.c_float, order='C' )
    return out

# def p_o_32(o, mus, sigmas, out=None):
#     # number of states
#     N = mus.shape[0]
#
#     pmus    = <float*> numpy.PyArray_DATA(mus)
#     psigmas = <float*> numpy.PyArray_DATA(sigmas)
#     p  = cdef_single_vector(N)
#     pp = <float*> numpy.PyArray_DATA(p)
#
#     return _p_o(o, pmus, psigmas, N, pp)


def p_o_64(o, mus, sigmas, out=None):
    # number of states
    N = mus.shape[0]

    pmus    = <double*> numpy.PyArray_DATA(mus)
    psigmas = <double*> numpy.PyArray_DATA(sigmas)
    if out is None:
        p = cdef_double_vector(N)
    else:
        p = out
    pp = <double*> numpy.PyArray_DATA(p)

    _p_o(o, pmus, psigmas, N, pp)

    return p


def p_o(o, mus, sigmas, out=None, dtype=numpy.float32):
    # check types
    assert(mus.dtype == dtype)
    assert(sigmas.dtype == dtype)

    # pointers to arrays
    if dtype == numpy.float32:
        raise ValueError
    elif dtype == numpy.float64:
        return p_o_64(o, mus, sigmas, out=out)
    else:
        raise TypeError


def p_obs_64(obs, mus, sigmas, out=None):
    N = mus.shape[0]
    T = obs.shape[0]
    pobs    = <double*> numpy.PyArray_DATA(obs)
    pmus    = <double*> numpy.PyArray_DATA(mus)
    psigmas = <double*> numpy.PyArray_DATA(sigmas)
    if out is None:
        p = cdef_double_matrix(T,N)
    else:
        p = out
    pp      = <double*> numpy.PyArray_DATA(p)

    _p_obs(pobs, pmus, psigmas, N, T, pp)

    return p


def p_obs(obs, mus, sigmas, out=None, dtype=numpy.float32):
    if (obs.dtype != dtype):
        obs = obs.astype(dtype)
    if (mus.dtype != dtype):
        mus = mus.astype(dtype)
    if (sigmas.dtype != dtype):
        sigmas = sigmas.astype(dtype)
    # check types
    assert(obs.dtype == dtype)
    assert(mus.dtype == dtype)
    assert(sigmas.dtype == dtype)

    # pointers to arrays
    if dtype == numpy.float32:
        raise ValueError
    elif dtype == numpy.float64:
        return p_obs_64(obs, mus, sigmas, out=out)
    else:
        raise TypeError



#!/usr/local/bin/env python

"""
Test HMM functions.

"""

import numpy as np
import unittest

import bhmm
from bhmm.util import testsystems

from nose.tools import assert_equal, assert_almost_equal
from numpy.testing import assert_array_almost_equal

__author__ = "John D. Chodera, Frank Noe"
__copyright__ = "Copyright 2015, John D. Chodera and Frank Noe"
__credits__ = ["John D. Chodera", "Frank Noe"]
__license__ = "FreeBSD"
__maintainer__ = "John D. Chodera"
__email__="jchodera AT gmail DOT com"

class TestHMM(unittest.TestCase):

    def test_hmm(self):
        # Create a simple HMM model.
        model = testsystems.dalton_model(nstates=3)
        # Test model parameter access.
        assert_equal(model.transition_matrix.shape, (3,3))
        assert_equal(model.stationary_distribution.shape, (3,))

        return

    def test_two_state_model(self):
        """Test the creation of a simple two-state HMM model with analytical parameters.
        """
        from bhmm import HMM
        # Create a simple two-state model.
        nstates = 2
        Tij = testsystems.generate_transition_matrix(reversible=True)
        from bhmm import GaussianOutputModel
        means=[-1,+1]
        sigmas=[1,1]
        output_model = GaussianOutputModel(nstates, means=means, sigmas=sigmas)
        model = bhmm.HMM(Tij, output_model)
        # Compute stationary probability using ARPACK.
        from scipy.sparse.linalg import eigs
        from numpy.linalg import norm
        [eigenvalues, eigenvectors] = eigs(Tij.T, k=1, which='LR')
        eigenvectors = np.real(eigenvectors)
        Pi = eigenvectors[:,0] / eigenvectors[:,0].sum()
        # Test model is correct.
        assert_array_almost_equal(model._Tij, Tij)
        assert_array_almost_equal(model._Pi, Pi)
        assert(np.allclose(model.output_model.means, np.array(means)))
        assert(np.allclose(model.output_model.sigmas, np.array(sigmas)))


if __name__=="__main__":
    unittest.main()



"""
Hidden Markov model

"""

__author__ = "Frank Noe and John D. Chodera"
__copyright__ = "Copyright 2015, John D. Chodera and Frank Noe"
__credits__ = ["Frank Noe", "John D. Chodera"]
__license__ = "LGPL"
__maintainer__ = "Frank Noe"
__email__="frank DOT noe AT fu-berlin DOT de"

import time
import copy

import numpy as np


# TODO: reactivate multiprocessing
# from multiprocessing import Queue, Process, cpu_count

# BHMM imports
import bhmm
import bhmm.hidden as hidden
from bhmm.util.logger import logger
from bhmm.util import config

class MaximumLikelihoodEstimator(object):
    """
    Maximum likelihood Hidden Markov model (HMM).

    This class is used to fit a maximum-likelihood HMM to data.

    Examples
    --------

    >>> import bhmm
    >>> bhmm.config.verbose = False
    >>>
    >>> from bhmm import testsystems
    >>> [model, O, S] = testsystems.generate_synthetic_observations(ntrajectories=5, length=1000)
    >>> mlhmm = MaximumLikelihoodEstimator(O, model.nstates)
    >>> model = mlhmm.fit()

    References
    ----------
    [1] L. E. Baum and J. A. Egon, "An inequality with applications to statistical estimation for probabilistic
        functions of a Markov process and to a model for ecology," Bull. Amer. Meteorol. Soc., vol. 73, pp. 360-363, 1967.

    """
    def __init__(self, observations, nstates, initial_model=None, type='gaussian',
                 reversible=True, stationary=True, p=None, accuracy=1e-3, maxit=1000):
        """Initialize a Bayesian hidden Markov model sampler.

        Parameters
        ----------
        observations : list of numpy arrays representing temporal data
            `observations[i]` is a 1d numpy array corresponding to the observed trajectory index `i`
        nstates : int
            The number of states in the model.
        initial_model : HMM, optional, default=None
            If specified, the given initial model will be used to initialize the BHMM.
            Otherwise, a heuristic scheme is used to generate an initial guess.
        type : str, optional, default=None
            Output model type from [None, 'gaussian', 'discrete'].
        reversible : bool, optional, default=True
            If True, a prior that enforces reversible transition matrices (detailed balance) is used;
            otherwise, a standard  non-reversible prior is used.
        stationary : bool, optional, default=True
            If True, the initial distribution of hidden states is self-consistently computed as the stationary
            distribution of the transition matrix. If False, it will be estimated from the starting states.
        p : ndarray (nstates), optional, default=None
            Initial or fixed stationary distribution. If given and stationary=True, transition matrices will be
            estimated with the constraint that they have p as their stationary distribution. If given and
            stationary=False, p is the fixed initial distribution of hidden states.
        accuracy : float
            convergence threshold for EM iteration. When two the likelihood does not increase by more than accuracy, the
            iteration is stopped successfully.
        maxit : int
            stopping criterion for EM iteration. When so many iterations are performanced without reaching the requested
            accuracy, the iteration is stopped without convergence (a warning is given)

        """
        # Store a copy of the observations.
        self._observations = copy.deepcopy(observations)
        self._nobs = len(observations)
        self._Ts = [len(o) for o in observations]
        self._maxT = np.max(self._Ts)

        # Store the number of states.
        self._nstates = nstates

        if initial_model is not None:
            # Use user-specified initial model, if provided.
            self._hmm = copy.deepcopy(initial_model)
            # consistency checks
            if self._hmm.is_stationary != stationary:
                logger().warn('Requested stationary=' + str(stationary) + ' but initial model is stationary=' +
                              str(self._hmm.is_stationary) + '. Using stationary='+str(self._hmm.is_stationary))
            if self._hmm.is_reversible != reversible:
                logger().warn('Requested reversible=' + str(reversible) + ' but initial model is reversible=' +
                              str(self._hmm.is_reversible) + '. Using reversible='+str(self._hmm.is_reversible))
            # setting parameters
            self._reversible = self._hmm.is_reversible
            self._stationary = self._hmm.is_stationary
        else:
            # Generate our own initial model.
            self._hmm = bhmm.init_hmm(observations, nstates, type=type)
            # setting parameters
            self._reversible = reversible
            self._stationary = stationary

        # stationary and initial distribution
        self._fixed_stationary_distribution = None
        self._fixed_initial_distribution = None
        if p is not None:
            if stationary:
                self._fixed_stationary_distribution = np.array(p)
            else:
                self._fixed_initial_distribution = np.array(p)

        # pre-construct hidden variables
        self._alpha = np.zeros((self._maxT, self._nstates), config.dtype, order='C')
        self._beta = np.zeros((self._maxT, self._nstates), config.dtype, order='C')
        self._pobs = np.zeros((self._maxT, self._nstates), config.dtype, order='C')
        self._gammas = [np.zeros((len(self._observations[i]), self._nstates), config.dtype, order='C')
                        for i in range(self._nobs)]
        self._Cs = [np.zeros((self._nstates, self._nstates), config.dtype, order='C') for _ in range(self._nobs)]

        # convergence options
        self._accuracy = accuracy
        self._maxit = maxit
        self._likelihoods = None

        # Kernel for computing things
        hidden.set_implementation(config.kernel)
        self._hmm.output_model.set_implementation(config.kernel)

    @property
    def observations(self):
        r""" Observation trajectories """
        return self._observations

    @property
    def nobservations(self):
        r""" Number of observation trajectories """
        return self._nobs

    @property
    def observation_lengths(self):
        r""" Lengths of observation trajectories """
        return self._Ts

    @property
    def is_reversible(self):
        r""" Whether the transition matrix is estimated with detailed balance constraints """
        return self._reversible

    @property
    def nstates(self):
        r""" Number of hidden states """
        return self._nstates

    @property
    def accuracy(self):
        r""" Convergence threshold for EM iteration """
        return self._accuracy

    @property
    def maxit(self):
        r""" Maximum number of iterations """
        return self._maxit

    @property
    def likelihood(self):
        r""" Estimated HMM likelihood """
        return self._likelihoods[-1]

    @property
    def likelihoods(self):
        r""" Sequence of likelihoods generated from the iteration """
        return self._likelihoods

    @property
    def hidden_state_probabilities(self):
        r""" Probabilities of hidden states at every trajectory and time point """
        return self._gammas

    @property
    def hmm(self):
        r""" The estimated HMM """
        return self._hmm

    @property
    def output_model(self):
        r""" The HMM output model """
        return self._hmm.output_model

    @property
    def transition_matrix(self):
        r""" Hidden transition matrix """
        return self._hmm.Tij

    @property
    def initial_probability(self):
        r""" Initial probability """
        return self._hmm.Pi

    @property
    def stationary_probability(self):
        r""" Stationary probability, if the model is stationary """
        assert self._stationary, 'Estimator is not stationary'
        return self._hmm.Pi

    def _forward_backward(self, itraj):
        """
        Estimation step: Runs the forward-back algorithm on trajectory with index itraj

        Parameters
        ----------
        itraj : int
            index of the observation trajectory to process

        Results
        -------
        logprob : float
            The probability to observe the observation sequence given the HMM
            parameters
        gamma : ndarray(T,N, dtype=float)
            state probabilities for each t
        count_matrix : ndarray(N,N, dtype=float)
            the Baum-Welch transition count matrix from the hidden state
            trajectory

        """
        # get parameters
        A = self._hmm.transition_matrix
        pi = self._hmm.stationary_distribution
        obs = self._observations[itraj]
        T = len(obs)
        # compute output probability matrix
        self._hmm.output_model.p_obs(obs, out=self._pobs)
        # forward variables
        logprob = hidden.forward(A, self._pobs, pi, T=T, alpha_out=self._alpha)[0]
        # backward variables
        hidden.backward(A, self._pobs, T=T, beta_out=self._beta)
        # gamma
        hidden.state_probabilities(self._alpha, self._beta, T=T, gamma_out=self._gammas[itraj])
        # count matrix
        hidden.transition_counts(self._alpha, self._beta, A, self._pobs, T=T, out=self._Cs[itraj])
        # return results
        return logprob

    def _update_model(self, gammas, count_matrices):
        """
        Maximization step: Updates the HMM model given the hidden state assignment and count matrices

        Parameters
        ----------
        gamma : [ ndarray(T,N, dtype=float) ]
            list of state probabilities for each trajectory
        count_matrix : [ ndarray(N,N, dtype=float) ]
            list of the Baum-Welch transition count matrices for each hidden state trajectory

        """
        K = len(self._observations)
        N = self._nstates

        C = np.zeros((N, N))
        gamma0_sum = np.zeros((N))
        for k in range(K):
            # update state counts
            gamma0_sum += gammas[k][0]
            # update count matrix
            C += count_matrices[k]

        logger().info("Count matrix = \n"+str(C))

        # compute new transition matrix
        from bhmm.estimators._tmatrix_disconnected import estimate_P, stationary_distribution
        T = estimate_P(C, reversible=self._hmm.is_reversible, fixed_statdist=self._fixed_stationary_distribution)
        # stationary or init distribution
        if self._hmm.is_stationary:
            if self._fixed_stationary_distribution is None:
                pi = stationary_distribution(C, T)
            else:
                pi = self._fixed_stationary_distribution
        else:
            if self._fixed_initial_distribution is None:
                pi = gamma0_sum / np.sum(gamma0_sum)
            else:
                pi = self._fixed_initial_distribution

        # update model
        self._hmm.update(T, pi)

        logger().info("T: \n"+str(T))
        logger().info("pi: \n"+str(pi))

        # update output model
        # TODO: need to parallelize model fitting. Otherwise we can't gain much speed!
        self._hmm.output_model._estimate_output_model(self._observations, gammas)

    def compute_viterbi_paths(self):
        """
        Computes the viterbi paths using the current HMM model

        """
        # get parameters
        K = len(self._observations)
        A = self._hmm.transition_matrix
        pi = self._hmm.stationary_distribution

        # compute viterbi path for each trajectory
        paths = np.empty((K), dtype=object)
        for itraj in range(K):
            obs = self._observations[itraj]
            # compute output probability matrix
            pobs = self._hmm.output_model.p_obs(obs)
            # hidden path
            paths[itraj] = hidden.viterbi(A, pobs, pi)

        # done
        return paths

    def fit(self):
        """
        Maximum-likelihood estimation of the HMM using the Baum-Welch algorithm

        Returns
        -------
        model : HMM
            The maximum likelihood HMM model.

        """
        logger().info("=================================================================")
        logger().info("Running Baum-Welch:")
        logger().info("  input observations: "+str(self.nobservations)+" of lengths "+str(self.observation_lengths))
        logger().info("  initial HMM guess:"+str(self._hmm))

        initial_time = time.time()

        it = 0
        self._likelihoods = np.zeros((self.maxit))
        loglik = 0.0
        converged = False

        while (not converged and it < self.maxit):
            loglik = 0.0
            for k in range(self._nobs):
                loglik += self._forward_backward(k)

            self._update_model(self._gammas, self._Cs)
            logger().info(str(it)+" ll = "+str(loglik))
            #print self.model.output_model
            #print "---------------------"

            self._likelihoods[it] = loglik

            if it > 0:
                if loglik - self._likelihoods[it-1] < self._accuracy:
                    #print "CONVERGED! Likelihood change = ",(loglik - self.likelihoods[it-1])
                    converged = True

            it += 1

        # truncate likelihood history
        self._likelihoods = self._likelihoods[:it]
        # set final likelihood
        self._hmm.likelihood = loglik

        final_time = time.time()
        elapsed_time = final_time - initial_time

        logger().info("maximum likelihood HMM:"+str(self._hmm))
        logger().info("Elapsed time for Baum-Welch solution: %.3f s" % elapsed_time)
        logger().info("Computing Viterbi path:")

        initial_time = time.time()

        # Compute hidden state trajectories using the Viterbi algorithm.
        self._hmm.hidden_state_trajectories = self.compute_viterbi_paths()

        final_time = time.time()
        elapsed_time = final_time - initial_time

        logger().info("Elapsed time for Viterbi path computation: %.3f s" % elapsed_time)
        logger().info("=================================================================")

        return self._hmm


    # TODO: reactive multiprocessing
    # ###################
    # # MULTIPROCESSING
    #
    # def _forward_backward_worker(self, work_queue, done_queue):
    #     try:
    #         for k in iter(work_queue.get, 'STOP'):
    #             (weight, gamma, count_matrix) = self._forward_backward(k)
    #             done_queue.put((k, weight, gamma, count_matrix))
    #     except Exception, e:
    #         done_queue.put(e.message)
    #     return True
    #
    #
    # def fit_parallel(self):
    #     """
    #     Maximum-likelihood estimation of the HMM using the Baum-Welch algorithm
    #
    #     Returns
    #     -------
    #     The hidden markov model
    #
    #     """
    #     K = len(self.observations)#, len(A), len(B[0])
    #     gammas = np.empty(K, dtype=object)
    #     count_matrices = np.empty(K, dtype=object)
    #
    #     it        = 0
    #     converged = False
    #
    #     num_threads = min(cpu_count(), K)
    #     work_queue = Queue()
    #     done_queue = Queue()
    #     processes = []
    #
    #     while (not converged):
    #         print "it", it
    #         loglik = 0.0
    #
    #         # fill work queue
    #         for k in range(K):
    #             work_queue.put(k)
    #
    #         # start processes
    #         for w in xrange(num_threads):
    #             p = Process(target=self._forward_backward_worker, args=(work_queue, done_queue))
    #             p.start()
    #             processes.append(p)
    #             work_queue.put('STOP')
    #
    #         # end processes
    #         for p in processes:
    #             p.join()
    #
    #         # done signal
    #         done_queue.put('STOP')
    #
    #         # get results
    #         for (k, ll, gamma, count_matrix) in iter(done_queue.get, 'STOP'):
    #             loglik += ll
    #             gammas[k] = gamma
    #             count_matrices[k] = count_matrix
    #
    #         # update T, pi
    #         self._update_model(gammas, count_matrices)
    #
    #         self.likelihoods[it] = loglik
    #
    #         if it > 0:
    #             if loglik - self.likelihoods[it-1] < self.accuracy:
    #                 converged = True
    #
    #         it += 1
    #
    #     return self.model



from __future__ import print_function
from six.moves import range
import numpy as np

from bhmm.output_models.impl_c import gaussian as gc
from bhmm.output_models import OutputModel
from bhmm.util.logger import logger
from bhmm.util import config

__author__ = "John D. Chodera, Frank Noe"
__copyright__ = "Copyright 2015, John D. Chodera and Frank Noe"
__credits__ = ["John D. Chodera", "Frank Noe"]
__license__ = "LGPL"
__maintainer__ = "John D. Chodera, Frank Noe"
__email__="jchodera AT gmail DOT com, frank DOT noe AT fu-berlin DOT de"

class GaussianOutputModel(OutputModel):
    """
    HMM output probability model using 1D-Gaussians

    """

    def __init__(self, nstates, means=None, sigmas=None):
        """
        Create a 1D Gaussian output model.

        Parameters
        ----------
        nstates : int
            The number of output states.
        means : array_like of shape (nstates,), optional, default=None
            If specified, initialize the Gaussian means to these values.
        sigmas : array_like of shape (nstates,), optional, default=None
            If specified, initialize the Gaussian variances to these values.

        Examples
        --------

        Create an observation model.

        >>> output_model = GaussianOutputModel(nstates=3, means=[-1, 0, 1], sigmas=[0.5, 1, 2])

        """
        OutputModel.__init__(self, nstates)

        dtype = config.dtype # type for internal storage

        if means is not None:
            self._means = np.array(means, dtype=dtype)
            if self._means.shape != (nstates,):
                raise Exception('means must have shape (%d,); instead got %s' % (nstates, str(self._means.shape)))
        else:
            self._means = np.zeros([nstates], dtype=dtype)

        if sigmas is not None:
            self._sigmas = np.array(sigmas, dtype=dtype)
            if self._sigmas.shape != (nstates,):
                raise Exception('sigmas must have shape (%d,); instead got %s' % (nstates, str(self._sigmas.shape)))
        else:
            self._sigmas = np.zeros([nstates], dtype=dtype)

        return

    def __repr__(self):
        r""" String representation of this output model
        >>> output_model = GaussianOutputModel(nstates=3, means=[-1, 0, 1], sigmas=[0.5, 1, 2])
        >>> print(repr(output_model))
        GaussianOutputModel(3, means=array([-1.,  0.,  1.]), sigmas=array([ 0.5,  1. ,  2. ]))

        """

        return "GaussianOutputModel(%d, means=%s, sigmas=%s)" % (self.nstates, repr(self.means), repr(self.sigmas))

    def __str__(self):
        r""" Human-readable string representation of this output model
        >>> output_model = GaussianOutputModel(nstates=3, means=[-1, 0, 1], sigmas=[0.5, 1, 2])
        >>> print(str(output_model))
        --------------------------------------------------------------------------------
        GaussianOutputModel
        nstates: 3
        means: [-1.  0.  1.]
        sigmas: [ 0.5  1.   2. ]
        --------------------------------------------------------------------------------
        """

        output  = "--------------------------------------------------------------------------------\n"
        output += "GaussianOutputModel\n"
        output += "nstates: %d\n" % self.nstates
        output += "means: %s\n" % str(self.means)
        output += "sigmas: %s\n" % str(self.sigmas)
        output += "--------------------------------------------------------------------------------"
        return output

    @property
    def model_type(self):
        r""" Model type. Returns 'gaussian' """
        return 'gaussian'

    @property
    def dimension(self):
        r""" Dimension of the Gaussian output model (currently 1) """
        return 1

    @property
    def means(self):
        r""" Mean values of Gaussians output densities """
        return self._means

    @property
    def sigmas(self):
        # TODO: Should we not rather give the variances? In the multidimensional case on usually uses the covariance
        # TODO:   matrix instead of its square root.
        r""" Standard deviations of Gaussian output densities """
        return self._sigmas

    # TODO: remove code when we're sure we don't need it
    # TODO: when cleaning up, save the functionality of the last function (log_p_o) and integrate into and overwrite of log_pobs.
    # def p_o_i(self, o, i):
    #     """
    #     Returns the output probability for symbol o given hidden state i
    #
    #     Parameters
    #     ----------
    #     o : float or array_like
    #         observation or observations for which probability is to be computed
    #     i : int
    #         the hidden state index
    #
    #     Return
    #     ------
    #     p_o_i : float
    #         the probability that hidden state i generates symbol o
    #
    #     Examples
    #     --------
    #
    #     Compute the output probability of a single observation from a given hidden state.
    #
    #     Create an observation model.
    #
    #     >>> output_model = GaussianOutputModel(nstates=3, means=[-1, 0, 1], sigmas=[0.5, 1, 2])
    #
    #     Compute the output probability of a single observation from a single state.
    #
    #     >>> observation = 0
    #     >>> state_index = 0
    #     >>> p_o_i = output_model.p_o_i(observation, state_index)
    #
    #     Compute the output probability of a vector of observations from a single state.
    #
    #     >>> observations = np.random.randn(100)
    #     >>> state_index = 0
    #     >>> p_o_i = output_model.p_o_i(observations, state_index)
    #
    #     """
    #     C = 1.0 / (np.sqrt(2.0 * np.pi) * self.sigmas[i])
    #     Pobs = C * np.exp(-0.5 * ((o - self.means[i]) / self.sigmas[i])**2)
    #     return Pobs
    #
    # def log_p_o_i(self, o, i):
    #     """
    #     Returns the log output probability for symbol o given hidden state i
    #
    #     Parameters
    #     ----------
    #     o : float or array_like
    #         observation or observations for which probability is to be computed
    #     i : int
    #         the hidden state index
    #
    #     Return
    #     ------
    #     log_p_o_i : float
    #         the probability that hidden state i generates symbol o
    #
    #     Examples
    #     --------
    #
    #     Compute the log output probability of a single observation from a given hidden state.
    #
    #     Create an observation model.
    #
    #     >>> output_model = GaussianOutputModel(nstates=3, means=[-1, 0, 1], sigmas=[0.5, 1, 2])
    #
    #     Compute the output probability of a single observation from a single state.
    #
    #     >>> observation = 0
    #     >>> state_index = 0
    #     >>> log_p_o_i = output_model.log_p_o_i(observation, state_index)
    #
    #     Compute the output probability of a vector of observations from a single state.
    #
    #     >>> observations = np.random.randn(100)
    #     >>> state_index = 0
    #     >>> p_o_i = output_model.p_o_i(observations, state_index)
    #
    #     """
    #     log_C = - 0.5 * np.log(2.0 * np.pi) - np.log(self.sigmas[i])
    #     log_Pobs = log_C - 0.5 * ((o - self.means[i]) / self.sigmas[i])**2
    #     return log_Pobs
    #
    #
    # def log_p_o(self, o):
    #     """
    #     Returns the log output probability for symbol o from all hidden states
    #
    #     Parameters
    #     ----------
    #     o : float
    #         A single observation.
    #
    #     Return
    #     ------
    #     log_p_o : ndarray (N)
    #         log_p_o[i] is the log probability density of the observation o from state i emission distribution
    #
    #     Examples
    #     --------
    #
    #     Create an observation model.
    #
    #     >>> output_model = GaussianOutputModel(nstates=3, means=[-1, 0, 1], sigmas=[0.5, 1, 2])
    #
    #     Compute the output probability of a single observation from all hidden states.
    #
    #     >>> observation = 0
    #     >>> log_p_o = output_model.log_p_o(observation)
    #
    #     """
    #     log_C = - 0.5 * np.log(2.0 * np.pi) - np.log(self.sigmas)
    #     log_Pobs = log_C - 0.5 * ((o-self.means)/self.sigmas)**2
    #     return log_Pobs

    def _p_o(self, o):
        """
        Returns the output probability for symbol o from all hidden states

        Parameters
        ----------
        o : float
            A single observation.

        Return
        ------
        p_o : ndarray (N)
            p_o[i] is the probability density of the observation o from state i emission distribution

        Examples
        --------

        Create an observation model.

        >>> output_model = GaussianOutputModel(nstates=3, means=[-1, 0, 1], sigmas=[0.5, 1, 2])

        Compute the output probability of a single observation from all hidden states.

        >>> observation = 0
        >>> p_o = output_model._p_o(observation)

        """
        if self.__impl__ == self.__IMPL_C__:
            return gc.p_o(o, self.means, self.sigmas, out=None, dtype=type(o))
        elif self.__impl__ == self.__IMPL_PYTHON__:
            C = 1.0 / (np.sqrt(2.0 * np.pi) * self.sigmas)
            Pobs = C * np.exp(-0.5 * ((o-self.means)/self.sigmas)**2)
            return Pobs
        else:
            raise RuntimeError('Implementation '+str(self.__impl__)+' not available')

    def p_obs(self, obs, out=None):
        """
        Returns the output probabilities for an entire trajectory and all hidden states

        Parameters
        ----------
        oobs : ndarray((T), dtype=int)
            a discrete trajectory of length T

        Return
        ------
        p_o : ndarray (T,N)
            the probability of generating the symbol at time point t from any of the N hidden states

        Examples
        --------

        Generate an observation model and synthetic observation trajectory.

        >>> nobs = 1000
        >>> output_model = GaussianOutputModel(nstates=3, means=[-1, 0, +1], sigmas=[0.5, 1, 2])
        >>> s_t = np.random.randint(0, output_model.nstates, size=[nobs])
        >>> o_t = output_model.generate_observation_trajectory(s_t)

        Compute output probabilities for entire trajectory and all hidden states.

        >>> p_o = output_model.p_obs(o_t)

        """
        if self.__impl__ == self.__IMPL_C__:
            return gc.p_obs(obs, self.means, self.sigmas, out=out, dtype=config.dtype)
        elif self.__impl__ == self.__IMPL_PYTHON__:
            T = len(obs)
            if out is None:
                res = np.zeros((T, self.nstates), dtype=config.dtype)
            else:
                res = out
            for t in range(T):
                res[t,:] = self._p_o(obs[t])
            return res
        else:
            raise RuntimeError('Implementation '+str(self.__impl__)+' not available')


    def _estimate_output_model(self, observations, weights):
        """
        Fits the output model given the observations and weights

        Parameters
        ----------
        observations : [ ndarray(T_k,) ] with K elements
            A list of K observation trajectories, each having length T_k and d dimensions
        weights : [ ndarray(T_k,nstates) ] with K elements
            A list of K weight matrices, each having length T_k
            weights[k][t,n] is the weight assignment from observations[k][t] to state index n

        Examples
        --------

        Generate an observation model and samples from each state.

        >>> ntrajectories = 3
        >>> nobs = 1000
        >>> output_model = GaussianOutputModel(nstates=3, means=[-1, 0, +1], sigmas=[0.5, 1, 2])
        >>> observations = [ np.random.randn(nobs) for _ in range(ntrajectories) ] # random observations
        >>> weights = [ np.random.dirichlet([2, 3, 4], size=nobs) for _ in range(ntrajectories) ] # random weights

        Update the observation model parameters my a maximum-likelihood fit.

        >>> output_model._estimate_output_model(observations, weights)

        """
        # sizes
        N = self.nstates
        K = len(observations)

        # fit means
        self._means = np.zeros((N))
        w_sum = np.zeros((N))
        for k in range(K):
            # update nominator
            for i in range(N):
                self.means[i] += np.dot(weights[k][:,i],observations[k])
            # update denominator
            w_sum += np.sum(weights[k], axis=0)
        # normalize
        self._means /= w_sum

        # fit variances
        self._sigmas  = np.zeros((N))
        w_sum = np.zeros((N))
        for k in range(K):
            # update nominator
            for i in range(N):
                Y = (observations[k]-self.means[i])**2
                self.sigmas[i] += np.dot(weights[k][:,i],Y)
            # update denominator
            w_sum += np.sum(weights[k], axis=0)
        # normalize
        self._sigmas /= w_sum
        self._sigmas = np.sqrt(self.sigmas)


    def _sample_output_model(self, observations):
        """
        Sample a new set of distribution parameters given a sample of observations from the given state.

        Both the internal parameters and the attached HMM model are updated.

        Parameters
        ----------
        observations :  [ numpy.array with shape (N_k,) ] with `nstates` elements
            observations[k] is a set of observations sampled from state `k`

        Examples
        --------

        Generate synthetic observations.

        >>> nstates = 3
        >>> nobs = 1000
        >>> output_model = GaussianOutputModel(nstates=nstates, means=[-1, 0, 1], sigmas=[0.5, 1, 2])
        >>> observations = [ output_model.generate_observations_from_state(state_index, nobs) for state_index in range(nstates) ]
        >>> weights = [ np.zeros([nobs,nstates], np.float32).T for _ in range(nstates) ]

        Update output parameters by sampling.

        >>> output_model._sample_output_model(observations)

        """
        for state_index in range(self.nstates):
            # Update state emission distribution parameters.

            observations_in_state = observations[state_index]
            # Determine number of samples in this state.
            nsamples_in_state = len(observations_in_state)

            # Skip update if no observations.
            if nsamples_in_state == 0:
                logger().warn('Warning: State %d has no observations.' % state_index)
                continue

            # Sample new mu.
            self.means[state_index] = np.random.randn()*self.sigmas[state_index]/np.sqrt(nsamples_in_state) + np.mean(observations_in_state)

            # Sample new sigma.
            # This scheme uses the improper Jeffreys prior on sigma^2, P(mu, sigma^2) \propto 1/sigma
            chisquared = np.random.chisquare(nsamples_in_state-1)
            sigmahat2 = np.mean((observations_in_state - self.means[state_index])**2)
            self.sigmas[state_index] = np.sqrt(sigmahat2) / np.sqrt(chisquared / nsamples_in_state)

        return


    def generate_observation_from_state(self, state_index):
        """
        Generate a single synthetic observation data from a given state.

        Parameters
        ----------
        state_index : int
            Index of the state from which observations are to be generated.

        Returns
        -------
        observation : float
            A single observation from the given state.

        Examples
        --------

        Generate an observation model.

        >>> output_model = GaussianOutputModel(nstates=2, means=[0, 1], sigmas=[1, 2])

        Generate sample from a state.

        >>> observation = output_model.generate_observation_from_state(0)

        """
        observation = self.sigmas[state_index] * np.random.randn() + self.means[state_index]
        return observation


    def generate_observations_from_state(self, state_index, nobs):
        """
        Generate synthetic observation data from a given state.

        Parameters
        ----------
        state_index : int
            Index of the state from which observations are to be generated.
        nobs : int
            The number of observations to generate.

        Returns
        -------
        observations : numpy.array of shape(nobs,)
            A sample of `nobs` observations from the specified state.

        Examples
        --------

        Generate an observation model.

        >>> output_model = GaussianOutputModel(nstates=2, means=[0, 1], sigmas=[1, 2])

        Generate samples from each state.

        >>> observations = [ output_model.generate_observations_from_state(state_index, nobs=100) for state_index in range(output_model.nstates) ]

        """
        observations = self.sigmas[state_index] * np.random.randn(nobs) + self.means[state_index]
        return observations


    def generate_observation_trajectory(self, s_t):
        """
        Generate synthetic observation data from a given state sequence.

        Parameters
        ----------
        s_t : numpy.array with shape (T,) of int type
            s_t[t] is the hidden state sampled at time t

        Returns
        -------
        o_t : numpy.array with shape (T,) of type dtype
            o_t[t] is the observation associated with state s_t[t]

        Examples
        --------

        Generate an observation model and synthetic state trajectory.

        >>> nobs = 1000
        >>> output_model = GaussianOutputModel(nstates=3, means=[-1, 0, +1], sigmas=[0.5, 1, 2])
        >>> s_t = np.random.randint(0, output_model.nstates, size=[nobs])

        Generate a synthetic trajectory

        >>> o_t = output_model.generate_observation_trajectory(s_t)

        """

        # Determine number of samples to generate.
        T = s_t.shape[0]

        o_t = np.zeros([T], dtype=config.dtype)
        for t in range(T):
            s = s_t[t]
            o_t[t] = self.sigmas[s] * np.random.randn() + self.means[s]
        return o_t





# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Free University
# Berlin, 14195 Berlin, Germany.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Created on Jul 25, 2014

@author: noe
'''

import numpy as np
import math
import itertools
from bhmm.util import types

def confidence_interval(data, alpha):
    """
    Computes the mean and alpha-confidence interval of the given sample set

    Parameters
    ----------
    data : ndarray
        a 1D-array of samples
    alpha : float in [0,1]
        the confidence level, i.e. percentage of data included in the interval
        
    Returns
    -------
    [m,l,r] where m is the mean of the data, and (l,r) are the m-alpha/2 and m+alpha/2 
    confidence interval boundaries.
    """
    if (alpha < 0 or alpha > 1):
        raise ValueError('Not a meaningful confidence level: '+str(alpha))
    
    # compute mean
    m = np.mean(data)
    # sort data
    sdata = np.sort(data)
    # index of the mean
    im = np.searchsorted(sdata, m)
    if (im == 0 or im == len(sdata)):
        pm = im
    else:
        pm = (im-1) + (m-sdata[im-1])/(sdata[im]-sdata[im-1])
    # left interval boundary
    pl = pm - alpha*(pm)
    il1 = max(0, int(math.floor(pl)))
    il2 = min(len(sdata)-1, int(math.ceil(pl)))
    l = sdata[il1] + (pl - il1)*(sdata[il2] - sdata[il1])
    # right interval boundary
    pr = pm + alpha*(len(data)-im)
    ir1 = max(0, int(math.floor(pr)))
    ir2 = min(len(sdata)-1, int(math.ceil(pr)))
    r = sdata[ir1] + (pr - ir1)*(sdata[ir2] - sdata[ir1])

    # return
    return (m, l, r)

def _indexes(arr):
    """
    Returns the list of all indexes of the given array. Currently works for one and two-dimensional arrays

    """
    myarr = np.array(arr)
    if myarr.ndim == 1:
        return range(len(myarr))
    elif myarr.ndim == 2:
        return itertools.product(range(arr.shape[0]), range(arr.shape[1]))
    else:
        raise NotImplementedError('Only supporting arrays of dimension 1 and 2 as yet.')

def _column(arr, indexes):
    """
    Returns a column with given indexes from a deep array

    For example, if the array is a matrix and indexes is a single int, will return arr[:,indexes].
    If the array is an order 3 tensor and indexes is a pair of ints, will return arr[:,indexes[0],indexes[1]], etc.

    """
    if arr.ndim == 2 and types.is_int(indexes):
        return arr[:,indexes]
    elif arr.ndim == 3 and len(indexes) == 2:
        return arr[:,indexes[0],indexes[1]]
    else:
        raise NotImplementedError('Only supporting arrays of dimension 2 and 3 as yet.')

def confidence_interval_arr(data, conf=0.95):
    r""" Computes element-wise confidence intervals from a sample of ndarrays

    Given a sample of arbitrarily shaped ndarrays, computes element-wise confidence intervals

    Parameters
    ----------
    data : ndarray (K, (shape))
        ndarray of ndarrays, the first index is a sample index, the remaining indexes are specific to the
        array of interest
    conf : float, optional, default = 0.95
        confidence interval

    Return
    ------
    lower : ndarray(shape)
        element-wise lower bounds
    upper : ndarray(shape)
        element-wise upper bounds

    """
    if (conf < 0 or conf > 1):
        raise ValueError('Not a meaningful confidence level: '+str(conf))

    # list or 1D-array? then fuse it
    if types.is_list(data) or (isinstance(data, np.ndarray) and np.ndim(data) == 1):
        newshape = tuple([len(data)] + list(data[0].shape))
        newdata = np.zeros(newshape)
        for i in range(len(data)):
            newdata[i,:] = data[i]
        data = newdata

    # do we have an array now? if yes go, if no fail
    if types.is_float_array(data):
        I = _indexes(data[0])
        lower = np.zeros(data[0].shape)
        upper = np.zeros(data[0].shape)
        for i in I:
            col = _column(data, i)
            m, lower[i], upper[i] = confidence_interval(col, conf)
        # return
        return (lower, upper)
    else:
        raise TypeError('data cannot be converted to an ndarray')



__author__ = 'noe'

import numpy as np

def estimate_P(C, reversible = True, fixed_statdist=None):
    # import emma
    import msmtools.estimation as msmest
    # output matrix. Initially eye
    n = np.shape(C)[0]
    P = np.eye((n), dtype=np.float64)
    # treat each connected set separately
    S = msmest.connected_sets(C)
    for s in S:
        if len(s) > 1: # if there's only one state, there's nothing to estimate and we leave it with diagonal 1
            # compute transition sub-matrix on s
            Cs = C[s,:][:,s]
            Ps = msmest.transition_matrix(Cs, reversible = reversible, mu=fixed_statdist)
            # write back to matrix
            for i,I in enumerate(s):
                for j,J in enumerate(s):
                    P[I,J] = Ps[i,j]
            P[s,:][:,s] = Ps
    # done
    return P

def stationary_distribution(C, P):
    # import emma
    import msmtools.estimation as msmest
    import msmtools.analysis as msmana
    # disconnected sets
    n = np.shape(C)[0]
    ctot = np.sum(C)
    pi = np.zeros((n))
    # treat each connected set separately
    S = msmest.connected_sets(C)
    for s in S:
        # compute weight
        w = np.sum(C[s,:]) / ctot
        pi[s] = w * msmana.statdist(P[s,:][:,s])
    # reinforce normalization
    pi /= np.sum(pi)
    return pi



"""
Bayesian hidden Markov models.

"""

import numpy as np
import copy
import time
#from scipy.misc import logsumexp
import bhmm.hidden as hidden
from bhmm.util.logger import logger
from bhmm.util import config

import msmtools.estimation as msmest
#from bhmm.msm.transition_matrix_sampling_rev import TransitionMatrixSamplerRev

__author__ = "John D. Chodera, Frank Noe"
__copyright__ = "Copyright 2015, John D. Chodera and Frank Noe"
__credits__ = ["John D. Chodera", "Frank Noe"]
__license__ = "LGPL"
__maintainer__ = "John D. Chodera"
__email__="jchodera AT gmail DOT com"

class BayesianHMMSampler(object):
    """Bayesian hidden Markov model sampler.

    Examples
    --------

    First, create some synthetic test data.

    >>> import bhmm
    >>> bhmm.config.verbose = False
    >>> nstates = 3
    >>> model = bhmm.testsystems.dalton_model(nstates)
    >>> [observations, hidden_states] = model.generate_synthetic_observation_trajectories(ntrajectories=5, length=1000)

    Initialize a new BHMM model.

    >>> from bhmm import BHMM
    >>> bhmm_sampler = BHMM(observations, nstates)

    Sample from the posterior.

    >>> models = bhmm_sampler.sample(nsamples=10)

    """
    def __init__(self, observations, nstates, initial_model=None,
                 reversible=True, transition_matrix_sampling_steps=1000, transition_matrix_prior=None,
                 type='gaussian'):
        """Initialize a Bayesian hidden Markov model sampler.

        Parameters
        ----------
        observations : list of numpy arrays representing temporal data
            `observations[i]` is a 1d numpy array corresponding to the observed trajectory index `i`
        nstates : int
            The number of states in the model.
        initial_model : HMM, optional, default=None
            If specified, the given initial model will be used to initialize the BHMM.
            Otherwise, a heuristic scheme is used to generate an initial guess.
        reversible : bool, optional, default=True
            If True, a prior that enforces reversible transition matrices (detailed balance) is used;
            otherwise, a standard  non-reversible prior is used.
        transition_matrix_sampling_steps : int, optional, default=1000
            number of transition matrix sampling steps per BHMM cycle
        transition_matrix_prior : str or ndarray(n,n)
            prior count matrix to be used for transition matrix sampling, or a keyword specifying the prior mode
            |  None (default),  -1 prior is used that ensures consistency between mean and MLE. Can lead to sampling
                disconnected matrices in the low-data regime. If you have disconnectivity problems, consider
                using 'init-connect'
            |  'init-connect',  prior count matrix ensuring the same connectivity as in the initial model. 1 count
                is added to all diagonals. All off-diagonals share one prior count distributed proportional to
                the row of the initial transition matrix.
        output_model_type : str, optional, default='gaussian'
            Output model type.  ['gaussian', 'discrete']

        """
        # Sanity checks.
        if len(observations) == 0:
            raise Exception("No observations were provided.")

        # Store options.
        self.reversible = reversible

        # Store the number of states.
        self.nstates = nstates

        # Store a copy of the observations.
        self.observations = copy.deepcopy(observations)
        self.nobs = len(observations)
        self.Ts = [len(o) for o in observations]
        self.maxT = np.max(self.Ts)

        # initial model
        if initial_model:
            # Use user-specified initial model, if provided.
            self.model = copy.deepcopy(initial_model)
        else:
            # Generate our own initial model.
            self.model = self._generateInitialModel(type)

        # prior counts
        if transition_matrix_prior is None:
            self.prior = np.zeros((self.nstates, self.nstates))
        elif isinstance(transition_matrix_prior, np.ndarray):
            if np.array_equal(transition_matrix_prior.shape, (self.nstates, self.nstates)):
                self.prior = np.array(transition_matrix_prior)
        elif transition_matrix_prior == 'init-connect':
            Pinit = self.model.transition_matrix
            self.prior = Pinit - np.diag(Pinit)  # add off-diagonals from initial T-matrix
            self.prior /= self.prior.sum(axis=1)[:, None]  # scale off-diagonals to row sum 1
            self.prior += np.eye(nstates)  # add diagonal 1.
        else:
            raise ValueError('transition matrix prior mode undefined: '+str(transition_matrix_prior))

        # sampling options
        self.transition_matrix_sampling_steps = transition_matrix_sampling_steps

        # implementation options
        hidden.set_implementation(config.kernel)
        self.model.output_model.set_implementation(config.kernel)

        # pre-construct hidden variables
        self.alpha = np.zeros((self.maxT,self.nstates), config.dtype, order='C')
        self.pobs = np.zeros((self.maxT,self.nstates), config.dtype, order='C')

        return

    def sample(self, nsamples, nburn=0, nthin=1, save_hidden_state_trajectory=False,
               call_back=None):
        """Sample from the BHMM posterior.

        Parameters
        ----------
        nsamples : int
            The number of samples to generate.
        nburn : int, optional, default=0
            The number of samples to discard to burn-in, following which `nsamples` will be generated.
        nthin : int, optional, default=1
            The number of Gibbs sampling updates used to generate each returned sample.
        save_hidden_state_trajectory : bool, optional, default=False
            If True, the hidden state trajectory for each sample will be saved as well.
        call_back : function, optional, default=None
            a call back function with no arguments, which if given is being called
            after each computed sample. This is useful for implementing progress bars.

        Returns
        -------
        models : list of bhmm.HMM
            The sampled HMM models from the Bayesian posterior.

        Examples
        --------

        >>> from bhmm import testsystems
        >>> [model, observations, states, sampled_model] = testsystems.generate_random_bhmm(ntrajectories=5, length=1000)
        >>> nburn = 5 # run the sampler a bit before recording samples
        >>> nsamples = 10 # generate 10 samples
        >>> nthin = 2 # discard one sample in between each recorded sample
        >>> samples = sampled_model.sample(nsamples, nburn=nburn, nthin=nthin)

        """

        # Run burn-in.
        for iteration in range(nburn):
            logger().info("Burn-in   %8d / %8d" % (iteration, nburn))
            self._update()

        # Collect data.
        models = list()
        for iteration in range(nsamples):
            logger().info("Iteration %8d / %8d" % (iteration, nsamples))
            # Run a number of Gibbs sampling updates to generate each sample.
            for thin in range(nthin):
                self._update()
            # Save a copy of the current model.
            model_copy = copy.deepcopy(self.model)
            #print "Sampled: \n",repr(model_copy)
            if not save_hidden_state_trajectory:
                model_copy.hidden_state_trajectory = None
            models.append(model_copy)
            if call_back is not None:
                call_back()

        # Return the list of models saved.
        return models

    def _update(self):
        """Update the current model using one round of Gibbs sampling.

        """
        initial_time = time.time()

        self._updateHiddenStateTrajectories()
        self._updateEmissionProbabilities()
        self._updateTransitionMatrix()

        final_time = time.time()
        elapsed_time = final_time - initial_time
        logger().info("BHMM update iteration took %.3f s" % elapsed_time)

    def _updateHiddenStateTrajectories(self):
        """Sample a new set of state trajectories from the conditional distribution P(S | T, E, O)

        """
        self.model.hidden_state_trajectories = list()
        for trajectory_index in range(self.nobs):
            hidden_state_trajectory = self._sampleHiddenStateTrajectory(self.observations[trajectory_index])
            self.model.hidden_state_trajectories.append(hidden_state_trajectory)
        return

    def _sampleHiddenStateTrajectory(self, obs, dtype=np.int32):
        """Sample a hidden state trajectory from the conditional distribution P(s | T, E, o)

        Parameters
        ----------
        o_t : numpy.array with dimensions (T,)
            observation[n] is the nth observation
        dtype : numpy.dtype, optional, default=numpy.int32
            The dtype to to use for returned state trajectory.

        Returns
        -------
        s_t : numpy.array with dimensions (T,) of type `dtype`
            Hidden state trajectory, with s_t[t] the hidden state corresponding to observation o_t[t]

        Examples
        --------
        >>> import bhmm
        >>> [model, observations, states, sampled_model] = bhmm.testsystems.generate_random_bhmm(ntrajectories=5, length=1000)
        >>> o_t = observations[0]
        >>> s_t = sampled_model._sampleHiddenStateTrajectory(o_t)

        """

        # Determine observation trajectory length
        T = obs.shape[0]

        # Convenience access.
        A = self.model.transition_matrix
        pi = self.model.initial_distribution

        # compute output probability matrix
        self.model.output_model.p_obs(obs, out=self.pobs)
        # forward variables
        logprob = hidden.forward(A, self.pobs, pi, T = T, alpha_out=self.alpha)[0]
        # sample path
        S = hidden.sample_path(self.alpha, A, self.pobs, T = T)

        return S

    def _updateEmissionProbabilities(self):
        """Sample a new set of emission probabilites from the conditional distribution P(E | S, O)

        """
        observations_by_state = [ self.model.collect_observations_in_state(self.observations, state) for state in range(self.model.nstates) ]
        self.model.output_model._sample_output_model(observations_by_state)
        return

    def _updateTransitionMatrix(self):
        """
        Updates the hidden-state transition matrix

        """
        C = self.model.count_matrix()
        # apply prior
        C += self.prior
        # sample T-matrix
        Tij = msmest.sample_tmatrix(C, nsample=1, nsteps=self.transition_matrix_sampling_steps, reversible=self.reversible)
        self.model.update(Tij)

    def _generateInitialModel(self, output_model_type):
        """Initialize using an MLHMM.

        """
        logger().info("Generating initial model for BHMM using MLHMM...")
        from bhmm.estimators.maximum_likelihood import MaximumLikelihoodEstimator
        mlhmm = MaximumLikelihoodEstimator(self.observations, self.nstates, reversible=self.reversible, type=output_model_type)
        model = mlhmm.fit()
        return model




from __future__ import print_function
__author__ = 'noe'

import numpy as np
import unittest
import time
from bhmm.output_models.gaussian import GaussianOutputModel

print_speedup = False

class TestOutputGaussian(unittest.TestCase):

    def setUp(self):
        nstates = 3
        means    = np.array([-0.5, 0.0, 0.5])
        sigmas = np.array([0.2, 0.2, 0.2])
        self.G = GaussianOutputModel(nstates, means=means, sigmas=sigmas)

        # random Gaussian samples
        self.obs = np.random.randn((10000))


    def tearDown(self):
        pass


    def test_p_obs(self):
        # compare results
        self.G.set_implementation('c')
        time1 = time.time()
        for i in range(10):
            p_c = self.G.p_obs(self.obs)
        time2 = time.time()
        t_c = time2-time1

        self.G.set_implementation('python')
        time1 = time.time()
        for i in range(10):
            p_p = self.G.p_obs(self.obs)
        time2 = time.time()
        t_p = time2-time1

        assert(np.allclose(p_c, p_p))

        # speed report
        if print_speedup:
            print('p_obs speedup c/python = '+str(t_p/t_c))



if __name__=="__main__":
    unittest.main()



from __future__ import print_function
__author__ = 'noe'

import unittest

import numpy as np
import time

from bhmm import hidden
from bhmm.output_models.gaussian import GaussianOutputModel

print_speedup = False

class TestHidden(unittest.TestCase):

    def setUp(self):
        self.nexamples = 0
        self.A    = []
        self.pi   = []
        self.pobs = []
        self.T    = []
        self.N    = []
        self.logprob = []
        self.alpha   = []
        self.time_alpha = []
        self.beta    = []
        self.time_beta = []
        self.gamma   = []
        self.time_gamma = []
        self.c       = []
        self.time_c = []
        self.C       = []
        self.time_C = []
        self.vpath   = []
        self.time_vpath = []
        self.alpha_mem   = []
        self.beta_mem   = []
        self.gamma_mem   = []
        self.C_mem   = []

        # first toy example
        A = np.array([[0.9, 0.1],
                      [0.1, 0.9]])
        pi = np.array([0.5, 0.5])
        pobs = np.array([[0.1, 0.9],
                         [0.1, 0.9],
                         [0.1, 0.9],
                         [0.1, 0.9],
                         [0.5, 0.5],
                         [0.9, 0.1],
                         [0.9, 0.1],
                         [0.9, 0.1],
                         [0.9, 0.1],
                         [0.9, 0.1]])
        self.append_example(A, pi, pobs)

        # second example
        A = np.array([[0.97, 0.02, 0.01],
                      [0.1,  0.8,  0.1],
                      [0.01, 0.02, 0.97]])
        pi = np.array([0.45, 0.1, 0.45])
        T = 10000
        means  = np.array([-1.0, 0.0, 1.0])
        sigmas = np.array([0.5,  0.5, 0.5])
        gom = GaussianOutputModel(3, means=means, sigmas=sigmas)
        obs = np.random.randint(3, size=T)
        pobs = gom.p_obs(obs)
        self.append_example(A, pi, pobs)


    def append_example(self, A, pi, pobs):
        i = len(self.A)
        self.A.append(A)
        self.pi.append(pi)
        self.pobs.append(pobs)
        self.N.append(A.shape[0])
        self.T.append(pobs.shape[0])
        (logprob, alpha, t) = self.run_forward(i, 'python', None)
        self.logprob.append(logprob)
        self.alpha.append(alpha)
        self.time_alpha.append(t)
        (beta, t) = self.run_backward(i, 'python', None)
        self.beta.append(beta)
        self.time_beta.append(t)
        (gamma, t) = self.run_gamma(i, 'python', None)
        self.gamma.append(gamma)
        self.time_gamma.append(t)
        (c, t) = self.run_state_counts(i, 'python', None)
        self.c.append(c)
        self.time_c.append(t)
        (C, t) = self.run_transition_counts(i, 'python', None)
        self.C.append(C)
        self.time_C.append(t)
        (vpath, t) = self.run_viterbi(i, 'python', None)
        self.vpath.append(vpath)
        self.time_vpath.append(t)
        #
        self.alpha_mem.append(np.zeros((pobs.shape[0],A.shape[0])))
        self.beta_mem.append(np.zeros((pobs.shape[0],A.shape[0])))
        self.gamma_mem.append(np.zeros((pobs.shape[0],A.shape[0])))
        self.C_mem.append(np.zeros((A.shape[0],A.shape[0])))

        self.nexamples += 1


    def run_all(self, A, pobs, pi):
        # forward
        logprob, alpha = hidden.forward(A, pobs, pi)
        # backward
        beta = hidden.backward(A, pobs)
        # gamma
        gamma = hidden.state_probabilities(alpha, beta)
        # state counts
        T = pobs.shape[0]
        statecount = hidden.state_counts(gamma, T)
        # transition counts
        C = hidden.transition_counts(alpha, beta, A, pobs)
        # viterbi path
        vpath = hidden.viterbi(A, pobs, pi)
        # return
        return (logprob, alpha, beta, gamma, statecount, C, vpath)


    def run_all_mem(self, A, pobs, pi):
        T = pobs.shape[0]
        N = A.shape[0]
        alpha = np.zeros( (T,N) )
        beta  = np.zeros( (T,N) )
        gamma = np.zeros( (T,N) )
        C     = np.zeros( (N,N) )
        logprob, alpha = hidden.forward(A, pobs, pi, alpha_out = alpha)
        # backward
        hidden.backward(A, pobs, beta_out = beta)
        # gamma
        hidden.state_probabilities(alpha, beta, gamma_out = gamma)
        # state counts
        statecount = hidden.state_counts(gamma, T)
        # transition counts
        hidden.transition_counts(alpha, beta, A, pobs, out=self.C)
        # viterbi path
        vpath = hidden.viterbi(A, pobs, pi)
        # return
        return (logprob, alpha, beta, gamma, statecount, C, vpath)


    def tearDown(self):
        pass


    def run_forward(self, i, kernel, out):
        nrep = max(1, int(10000/self.T[i]))
        logprob = 0
        alpha = None
        hidden.set_implementation(kernel)
        time1 = time.time()
        for k in range(nrep):
            logprob, alpha = hidden.forward(self.A[i], self.pobs[i], self.pi[i], alpha_out=out)
        # compare
        time2 = time.time()
        d = (time2-time1)/(1.0*nrep)
        return (logprob, alpha, d)

    def run_backward(self, i, kernel, out):
        nrep = max(1, int(10000/self.T[i]))
        beta = None
        hidden.set_implementation(kernel)
        time1 = time.time()
        for k in range(nrep):
            beta = hidden.backward(self.A[i], self.pobs[i], beta_out=out)
        # compare
        time2 = time.time()
        d = (time2-time1)/(1.0*nrep)
        return (beta, d)

    def run_gamma(self, i, kernel, out):
        nrep = max(1, int(10000/self.T[i]))
        gamma = None
        hidden.set_implementation(kernel)
        time1 = time.time()
        for k in range(nrep):
            gamma = hidden.state_probabilities(self.alpha[i], self.beta[i], gamma_out=out)
        # compare
        time2 = time.time()
        d = (time2-time1)/(1.0*nrep)
        return (gamma, d)

    def run_state_counts(self, i, kernel, out):
        nrep = max(1, int(10000/self.T[i]))
        c = None
        hidden.set_implementation(kernel)
        time1 = time.time()
        for k in range(nrep):
            c = hidden.state_counts(self.gamma[i], self.T[i])
        # compare
        time2 = time.time()
        d = (time2-time1)/(1.0*nrep)
        return (c, d)

    def run_transition_counts(self, i, kernel, out):
        nrep = max(1, int(10000/self.T[i]))
        C = None
        hidden.set_implementation(kernel)
        time1 = time.time()
        for k in range(nrep):
            C = hidden.transition_counts(self.alpha[i], self.beta[i], self.A[i], self.pobs[i], out=out)
        # compare
        time2 = time.time()
        d = (time2-time1)/(1.0*nrep)
        return (C, d)

    def run_viterbi(self, i, kernel, out):
        nrep = max(1, int(10000/self.T[i]))
        vpath = None
        hidden.set_implementation(kernel)
        time1 = time.time()
        for k in range(nrep):
            vpath = hidden.viterbi(self.A[i], self.pobs[i], self.pi[i])
        # compare
        time2 = time.time()
        d = (time2-time1)/(1.0*nrep)
        return (vpath, d)


    def run_abs(self, call, kernel):
        """
        Reference. Just computes the time
        """
        for i in range(self.nexamples):
            res = call(i, kernel, None)
            if (print_speedup):
                print('\t'+str(call.__name__)+'\t Example '+str(i)+'\t Impl = '+str(kernel)+' Time = '+str(res[-1]))

    def run_comp(self, call, kernel, outs, refs, reftime):
        """
        Reference. Just computes the time
        """
        for i in range(self.nexamples):
            if (outs is None):
                res = call(i, kernel, None)
            else:
                res = call(i, kernel, outs[i])
            for j in range(len(res)-1):
                myres  = res[j]
                refres = refs[j][i]
                self.assertTrue(np.allclose(myres, refres))
            if outs is None:
                pkernel = kernel
            else:
                pkernel = kernel + ' mem'
            if (print_speedup):
                print('\t'+str(call.__name__)+'\t Example '+str(i)+'\t Impl = '+pkernel+' Speedup = '+str(reftime[i]/res[-1]))


    def test_forward_p(self):
        self.run_abs(self.run_forward, 'python')

    def test_forward_p_mem(self):
        self.run_comp(self.run_forward, 'python', self.alpha_mem, [self.logprob, self.alpha], self.time_alpha)

    def test_forward_c(self):
        self.run_comp(self.run_forward, 'c', None, [self.logprob, self.alpha], self.time_alpha)

    def test_forward_c_mem(self):
        self.run_comp(self.run_forward, 'c', self.alpha_mem, [self.logprob, self.alpha], self.time_alpha)

    def test_backward_p(self):
        self.run_abs(self.run_backward, 'python')

    def test_backward_p_mem(self):
        self.run_comp(self.run_backward, 'python', self.beta_mem, [self.beta], self.time_beta)

    def test_backward_c(self):
        self.run_comp(self.run_backward, 'c', None, [self.beta], self.time_beta)

    def test_backward_c_mem(self):
        self.run_comp(self.run_backward, 'c', self.beta_mem, [self.beta], self.time_beta)

    def test_gamma_p(self):
        self.run_abs(self.run_gamma, 'python')

    def test_gamma_p_mem(self):
        self.run_comp(self.run_gamma, 'python', self.gamma_mem, [self.gamma], self.time_gamma)

    def test_gamma_c(self):
        self.run_comp(self.run_gamma, 'c', None, [self.gamma], self.time_gamma)

    def test_gamma_c_mem(self):
        self.run_comp(self.run_gamma, 'c', self.gamma_mem, [self.gamma], self.time_gamma)

    def test_state_counts_p(self):
        self.run_abs(self.run_state_counts, 'python')

    def test_state_counts_p_mem(self):
        self.run_comp(self.run_state_counts, 'python', None, [self.c], self.time_c)

    def test_state_counts_c(self):
        self.run_comp(self.run_state_counts, 'c', None, [self.c], self.time_c)

    def test_state_counts_c_mem(self):
        self.run_comp(self.run_state_counts, 'c', None, [self.c], self.time_c)

    def test_transition_counts_p(self):
        self.run_abs(self.run_transition_counts, 'python')

    def test_transition_counts_p_mem(self):
        self.run_comp(self.run_transition_counts, 'python', self.C_mem, [self.C], self.time_C)

    def test_transition_counts_c(self):
        self.run_comp(self.run_transition_counts, 'c', None, [self.C], self.time_C)

    def test_transition_counts_c_mem(self):
        self.run_comp(self.run_transition_counts, 'c', self.C_mem, [self.C], self.time_C)

    def test_viterbi_p(self):
        self.run_abs(self.run_viterbi, 'python')

    def test_viterbi_p_mem(self):
        self.run_comp(self.run_viterbi, 'python', None, [self.vpath], self.time_vpath)

    def test_viterbi_c(self):
        self.run_comp(self.run_viterbi, 'c', None, [self.vpath], self.time_vpath)

    def test_viterbi_c_mem(self):
        self.run_comp(self.run_viterbi, 'c', None, [self.vpath], self.time_vpath)

    def test_fbtime_p_mem(self):
        for i in range(self.nexamples):
            ttot = 0.0
            (logprob, alpha, t) = self.run_forward(i, 'python', self.alpha_mem[i])
            ttot += t
            (beta, t) = self.run_backward(i, 'python', self.beta_mem[i])
            ttot += t
            (gamma, t) = self.run_gamma(i, 'python', self.gamma_mem[i])
            ttot += t
            (c, t) = self.run_state_counts(i, 'python', None)
            ttot += t
            (C, t) = self.run_transition_counts(i, 'python', self.C_mem[i])
            ttot += t
            tref = self.time_alpha[i] + self.time_beta[i] + self.time_gamma[i] + self.time_c[i] + self.time_C[i]
            if (print_speedup):
                print ('TOTAL speedup forward-backward example '+str(i)+'\t impl=python mem: \t'+str(tref/ttot))

    def test_fbtime_c(self):
        for i in range(self.nexamples):
            ttot = 0.0
            (logprob, alpha, t) = self.run_forward(i, 'c', None)
            ttot += t
            (beta, t) = self.run_backward(i, 'c', None)
            ttot += t
            (gamma, t) = self.run_gamma(i, 'c', None)
            ttot += t
            (c, t) = self.run_state_counts(i, 'c', None)
            ttot += t
            (C, t) = self.run_transition_counts(i, 'c', None)
            ttot += t
            tref = self.time_alpha[i] + self.time_beta[i] + self.time_gamma[i] + self.time_c[i] + self.time_C[i]
            if (print_speedup):
                print ('TOTAL speedup forward-backward example '+str(i)+'\t impl=c: \t'+str(tref/ttot))

    def test_fbtime_c_mem(self):
        for i in range(self.nexamples):
            ttot = 0.0
            (logprob, alpha, t) = self.run_forward(i, 'c', self.alpha_mem[i])
            ttot += t
            (beta, t) = self.run_backward(i, 'c', self.beta_mem[i])
            ttot += t
            (gamma, t) = self.run_gamma(i, 'c', self.gamma_mem[i])
            ttot += t
            (c, t) = self.run_state_counts(i, 'c', None)
            ttot += t
            (C, t) = self.run_transition_counts(i, 'c', self.C_mem[i])
            ttot += t
            tref = self.time_alpha[i] + self.time_beta[i] + self.time_gamma[i] + self.time_c[i] + self.time_C[i]
            if (print_speedup):
                print ('TOTAL speedup forward-backward example '+str(i)+'\t impl=c mem: \t'+str(tref/ttot))



if __name__=="__main__":
    unittest.main()



__author__ = 'noe'

import numpy as _np

def _guess_model_type(observations):
    """ Suggests a HMM model type based on the observation data

    Uses simple rules in order to decide which HMM model type makes sense based on observation data.
    If observations consist of arrays/lists of integer numbers (irrespective of whether the python type is
    int or float), our guess is 'discrete'.
    If observations consist of arrays/lists of 1D-floats, our guess is 'discrete'.
    In any other case, a TypeError is raised because we are not supporting that data type yet.

    Parameters:
    -----------
    observations : list of lists or arrays
        observation trajectories

    Returns:
    --------
    model_type : str
        One of {'discrete', 'gaussian'}

    """
    from bhmm.util import types as _types

    o1 = _np.array(observations[0])

    # CASE: vector of int? Then we want a discrete HMM
    if _types.is_int_vector(o1):
        return 'discrete'

    # CASE: not int type, but everything is an integral number. Then we also go for discrete
    if _np.allclose(o1, _np.round(o1)):
        isintegral = True
        for i in range(1, len(observations)):
            if not _np.allclose(observations[i], _np.round(observations[i])):
                isintegral = False
                break
        if isintegral:
            return 'discrete'

    # CASE: vector of double? Then we want a gaussian
    if _types.is_float_vector(o1):
        return 'gaussian'

    # None of the above? Then we currently do not support this format!
    raise TypeError('Observations is neither sequences of integers nor 1D-sequences of floats. The current version'
                    'does not support your input.')

def _lag_observations(observations, lag):
    """ Create new trajectories that are subsampled at lag but shifted

    Given a trajectory (s0, s1, s2, s3, s4, ...) and lag 3, this function will generate 3 trajectories
    (s0, s3, s6, ...), (s1, s4, s7, ...) and (s2, s5, s8, ...). Use this function in order to parametrize a MLE
    at lag times larger than 1 without discarding data. Do not use this function for Bayesian estimators, where
    data must be given such that subsequent transitions are uncorrelated.

    """
    obsnew = []
    for obs in observations:
        for shift in range(0, lag):
            obsnew.append(obs[shift:][::lag])
    return obsnew

def init_hmm(observations, nstates, lag=1, type=None):
    """Use a heuristic scheme to generate an initial model.

    Parameters
    ----------
    observations : list of ndarray((T_i))
        list of arrays of length T_i with observation data
    nstates : int
        The number of states.
    type : str, optional, default=None
        Output model type from [None, 'gaussian', 'discrete']. If None, will automatically select an output
        model type based on the format of observations.

    Examples
    --------

    Generate initial model for a gaussian output model.

    >>> import bhmm
    >>> [model, observations, states] = bhmm.testsystems.generate_synthetic_observations(output_model_type='gaussian')
    >>> initial_model = init_hmm(observations, model.nstates, type='gaussian')

    Generate initial model for a discrete output model.

    >>> import bhmm
    >>> [model, observations, states] = bhmm.testsystems.generate_synthetic_observations(output_model_type='discrete')
    >>> initial_model = init_hmm(observations, model.nstates, type='discrete')

    """
    # select output model type
    if (type is None):
        type = _guess_model_type(observations)

    if type == 'discrete':
        from bhmm.init import discrete
        return discrete.initial_model_discrete(observations, nstates, lag=lag, reversible=True)
    elif type == 'gaussian':
        from bhmm.init import gaussian
        return gaussian.initial_model_gaussian1d(observations, nstates, reversible=True)
    else:
        raise NotImplementedError('output model type '+str(type)+' not yet implemented.')

def gaussian_hmm(P, means, sigmas, pi=None, stationary=True, reversible=True):
    """ Initializes a 1D-Gaussian HMM

    Parameters
    ----------
    P : ndarray(nstates,nstates)
        Hidden transition matrix
    means : ndarray(nstates, )
        Means of Gaussian output distributions
    sigmas : ndarray(nstates, )
        Standard deviations of Gaussian output distributions
    pi : ndarray(nstates, )
        Fixed initial (if stationary=False) or fixed stationary distribution (if stationary=True).
    stationary : bool, optional, default=True
        If True: initial distribution is equal to stationary distribution of transition matrix
    reversible : bool, optional, default=True
        If True: transition matrix will fulfill detailed balance constraints.

    """
    from bhmm.hmm.gaussian_hmm import GaussianHMM
    from bhmm.output_models.gaussian import GaussianOutputModel
    # count states
    nstates = _np.array(P).shape[0]
    # initialize output model
    output_model = GaussianOutputModel(nstates, means, sigmas)
    # initialize general HMM
    from bhmm.hmm.generic_hmm import HMM as _HMM
    ghmm = _HMM(P, output_model, Pi=pi, stationary=stationary, reversible=reversible)
    # turn it into a Gaussian HMM
    ghmm = GaussianHMM(ghmm)
    return ghmm

def discrete_hmm(P, pout, pi=None, stationary=True, reversible=True):
    """ Initializes a discrete HMM

    Parameters
    ----------
    P : ndarray(nstates,nstates)
        Hidden transition matrix
    pout : ndarray(nstates,nsymbols)
        Output matrix from hidden states to observable symbols
    pi : ndarray(nstates, )
        Fixed initial (if stationary=False) or fixed stationary distribution (if stationary=True).
    stationary : bool, optional, default=True
        If True: initial distribution is equal to stationary distribution of transition matrix
    reversible : bool, optional, default=True
        If True: transition matrix will fulfill detailed balance constraints.

    """
    from bhmm.hmm.discrete_hmm import DiscreteHMM
    from bhmm.output_models.discrete import DiscreteOutputModel
    # initialize output model
    output_model = DiscreteOutputModel(pout)
    # initialize general HMM
    from bhmm.hmm.generic_hmm import HMM as _HMM
    dhmm = _HMM(P, output_model, Pi=pi, stationary=stationary, reversible=reversible)
    # turn it into a Gaussian HMM
    dhmm = DiscreteHMM(dhmm)
    return dhmm

def estimate_hmm(observations, nstates, lag=1, initial_model=None, type=None,
                 reversible=True, stationary=True, p=None, accuracy=1e-3, maxit=1000):
    r""" Estimate maximum-likelihood HMM

    Generic maximum-likelihood estimation of HMMs

    Parameters
    ----------
    observations : list of numpy arrays representing temporal data
        `observations[i]` is a 1d numpy array corresponding to the observed trajectory index `i`
    nstates : int
        The number of states in the model.
    lag : int
        the lag time at which observations should be read
    initial_model : HMM, optional, default=None
        If specified, the given initial model will be used to initialize the BHMM.
        Otherwise, a heuristic scheme is used to generate an initial guess.
    type : str, optional, default=None
        Output model type from [None, 'gaussian', 'discrete']. If None, will automatically select an output
        model type based on the format of observations.
    reversible : bool, optional, default=True
        If True, a prior that enforces reversible transition matrices (detailed balance) is used;
        otherwise, a standard  non-reversible prior is used.
    stationary : bool, optional, default=True
        If True, the initial distribution of hidden states is self-consistently computed as the stationary
        distribution of the transition matrix. If False, it will be estimated from the starting states.
    p : ndarray (nstates), optional, default=None
        Initial or fixed stationary distribution. If given and stationary=True, transition matrices will be
        estimated with the constraint that they have p as their stationary distribution. If given and
        stationary=False, p is the fixed initial distribution of hidden states.
    accuracy : float
        convergence threshold for EM iteration. When two the likelihood does not increase by more than accuracy, the
        iteration is stopped successfully.
    maxit : int
        stopping criterion for EM iteration. When so many iterations are performed without reaching the requested
        accuracy, the iteration is stopped without convergence (a warning is given)

    Return
    ------
    hmm : :class:`HMM <bhmm.hmm.generic_hmm.HMM>`

    """
    # select output model type
    if (type is None):
        type = _guess_model_type(observations)

    if lag > 1:
        observations = _lag_observations(observations, lag)

    # construct estimator
    from bhmm.estimators.maximum_likelihood import MaximumLikelihoodEstimator as _MaximumLikelihoodEstimator
    est = _MaximumLikelihoodEstimator(observations, nstates, initial_model=initial_model, type=type,
                                      reversible=reversible, stationary=stationary, p=p, accuracy=accuracy, maxit=maxit)
    # run
    est.fit()
    # set lag time
    est.hmm._lag = lag
    # return model
    return est.hmm

def bayesian_hmm(observations, estimated_hmm, nsample=100, transition_matrix_prior=None,
                 store_hidden=False, call_back=None):
    r""" Bayesian HMM based on sampling the posterior

    Generic maximum-likelihood estimation of HMMs

    Parameters
    ----------
    observations : list of numpy arrays representing temporal data
        `observations[i]` is a 1d numpy array corresponding to the observed trajectory index `i`
    estimated_hmm : HMM
        HMM estimated from estimate_hmm or initialize_hmm
    nsample : int, optional, default=100
        number of Gibbs sampling steps
    transition_matrix_prior : str or ndarray(n,n)
        prior count matrix to be used for transition matrix sampling, or a keyword specifying the prior mode
        |  None (default),  -1 prior is used that ensures consistency between mean and MLE. Can lead to sampling
            disconnected matrices in the low-data regime. If you have disconnectivity problems, consider
            using 'init-connect'
        |  'init-connect',  prior count matrix ensuring the same connectivity as in the initial model. 1 count
            is added to all diagonals. All off-diagonals share one prior count distributed proportional to
            the row of the initial transition matrix.
    store_hidden : bool, optional, default=False
        store hidden trajectories in sampled HMMs
    call_back : function, optional, default=None
        a call back function with no arguments, which if given is being called
        after each computed sample. This is useful for implementing progress bars.

    Return
    ------
    hmm : :class:`SampledHMM <bhmm.hmm.generic_sampled_hmm.SampledHMM>`

    """
    # construct estimator
    from bhmm.estimators.bayesian_sampling import BayesianHMMSampler as _BHMM
    sampler = _BHMM(observations, estimated_hmm.nstates, initial_model=estimated_hmm,
                    reversible=estimated_hmm.is_reversible, transition_matrix_sampling_steps=1000,
                    transition_matrix_prior=transition_matrix_prior, type=estimated_hmm.output_model.model_type)

    # Sample models.
    sampled_hmms = sampler.sample(nsamples=nsample, save_hidden_state_trajectory=store_hidden,
                                  call_back=call_back)
    # return model
    from bhmm.hmm.generic_sampled_hmm import SampledHMM
    return SampledHMM(estimated_hmm, sampled_hmms)



__author__ = 'noe'

import numpy as np

from bhmm.hmm.generic_hmm import HMM
from bhmm.util.logger import logger
from bhmm.util import config

def initial_model_gaussian1d(observations, nstates, reversible=True):
    """Generate an initial model with 1D-Gaussian output densities

    Parameters
    ----------
    observations : list of ndarray((T_i), dtype=float)
        list of arrays of length T_i with observation data
    nstates : int
        The number of states.

    Examples
    --------

    Generate initial model for a gaussian output model.

    >>> from bhmm import testsystems
    >>> [model, observations, states] = testsystems.generate_synthetic_observations(output_model_type='gaussian')
    >>> initial_model = initial_model_gaussian1d(observations, model.nstates)

    """
    ntrajectories = len(observations)

    # Concatenate all observations.
    collected_observations = np.array([], dtype=config.dtype)
    for o_t in observations:
        collected_observations = np.append(collected_observations, o_t)

    # Fit a Gaussian mixture model to obtain emission distributions and state stationary probabilities.
    from bhmm._external.sklearn import mixture
    gmm = mixture.GMM(n_components=nstates)
    gmm.fit(collected_observations[:,None])
    from bhmm import GaussianOutputModel
    output_model = GaussianOutputModel(nstates, means=gmm.means_[:,0], sigmas=np.sqrt(gmm.covars_[:,0]))

    logger().info("Gaussian output model:\n"+str(output_model))

    # Extract stationary distributions.
    Pi = np.zeros([nstates], np.float64)
    Pi[:] = gmm.weights_[:]

    logger().info("GMM weights: %s" % str(gmm.weights_))

    # Compute fractional state memberships.
    Nij = np.zeros([nstates, nstates], np.float64)
    for o_t in observations:
        # length of trajectory
        T = o_t.shape[0]
        # output probability
        pobs = output_model.p_obs(o_t)
        # normalize
        pobs /= pobs.sum(axis=1)[:,None]
        # Accumulate fractional transition counts from this trajectory.
        for t in range(T-1):
            Nij[:,:] = Nij[:,:] + np.outer(pobs[t,:], pobs[t+1,:])

        logger().info("Nij\n"+str(Nij))

    # Compute transition matrix maximum likelihood estimate.
    import msmtools.estimation as msmest
    Tij = msmest.transition_matrix(Nij, reversible=reversible)

    # Update model.
    model = HMM(Tij, output_model, reversible=reversible)

    return model



""" BHMM: A toolkit for Bayesian hidden Markov model analysis of single-molecule trajectories.

This project provides tools for estimating the number of metastable states, rate
constants between the states, equilibrium populations, distributions
characterizing the states, and distributions of these quantities from
single-molecule data. This data could be FRET data, single-molecule pulling
data, or any data where one or more observables are recorded as a function of
time. A Hidden Markov Model (HMM) is used to interpret the observed dynamics,
and a distribution of models that fit the data is sampled using Bayesian
inference techniques and Markov chain Monte Carlo (MCMC), allowing for both the
characterization of uncertainties in the model and modeling of the expected
information gain by new experiments.
"""

from __future__ import print_function
__requires__ = ['setuptools>=18',]
import os
from os.path import relpath, join

import versioneer
import pkg_resources
from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext as _build_ext

DOCLINES = __doc__.split("\n")

########################
CLASSIFIERS = """\
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)
Programming Language :: Python
Topic :: Scientific/Engineering :: Bio-Informatics
Topic :: Scientific/Engineering :: Chemistry
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""

################################################################################
# USEFUL SUBROUTINES
################################################################################
def find_package_data(data_root, package_root):
    files = []
    for root, dirnames, filenames in os.walk(data_root):
        for fn in filenames:
            files.append(relpath(join(root, fn), package_root))
    return files


################################################################################
# SETUP
################################################################################

extensions = [Extension('bhmm.hidden.impl_c.hidden',
                        sources = ['./bhmm/hidden/impl_c/hidden.pyx',
                                   './bhmm/hidden/impl_c/_hidden.c'],
                        include_dirs = ['/bhmm/hidden/impl_c/']),
              Extension('bhmm.output_models.impl_c.discrete',
                        sources = ['./bhmm/output_models/impl_c/discrete.pyx',
                                   './bhmm/output_models/impl_c/_discrete.c'],
                        include_dirs = ['/bhmm/output_models/impl_c/']),
              Extension('bhmm.output_models.impl_c.gaussian',
                        sources = ['./bhmm/output_models/impl_c/gaussian.pyx',
                                   './bhmm/output_models/impl_c/_gaussian.c'],
                        include_dirs = ['/bhmm/output_models/impl_c/']),
              Extension('bhmm._external.clustering.kmeans_clustering',
                        sources=['./bhmm/_external/clustering/src/clustering.c',
                                 './bhmm/_external/clustering/src/kmeans.c'],
                        include_dirs=['./bhmm/_external/clustering/include',
                                     ],
                        extra_compile_args=['-std=c99']),
              ]


class build_ext(_build_ext):
    def initialize_options(self):
        _build_ext.initialize_options(self)
        import pkg_resources
        dir = pkg_resources.resource_filename('numpy', 'core/include')
        self.include_dirs = [dir]

cmdclass = versioneer.get_cmdclass()
cmdclass['build_ext'] = build_ext

setup(
    name='bhmm',
    author='John Chodera and Frank Noe',
    author_email='john.chodera@choderalab.org',
    description=DOCLINES[0],
    long_description="\n".join(DOCLINES[2:]),
    version=versioneer.get_version(),
    cmdclass=cmdclass,
    license='LGPL',
    url='https://github.com/bhmm/bhmm',
    platforms=['Linux', 'Mac OS-X', 'Unix', 'Windows'],
    classifiers=CLASSIFIERS.splitlines(),
    package_dir={'bhmm': 'bhmm'},
    packages=find_packages(),
    package_data={'bhmm': find_package_data('examples', 'bhmm') + find_package_data('bhmm/tests/data', 'bhmm')},  # NOTE: examples installs to bhmm.egg/examples/, NOT bhmm.egg/bhmm/examples/.  You need to do utils.get_data_filename("../examples/*/setup/").
    zip_safe=False,
    install_requires=[
        'numpy',
        'scipy',
        'msmtools',
        'six',
        ],
    setup_requires=[
        'cython',
        'numpy',
        ],
    ext_modules=extensions,
    )




"""
Gaussian hidden Markov models

"""

__author__ = "John D. Chodera, Frank Noe"
__copyright__ = "Copyright 2015, John D. Chodera and Frank Noe"
__credits__ = ["John D. Chodera", "Frank Noe"]
__license__ = "LGPL"
__maintainer__ = "John D. Chodera"
__email__="jchodera AT gmail DOT com"

import numpy as np

from bhmm.hmm.generic_hmm import HMM
from bhmm.hmm.generic_sampled_hmm import SampledHMM
from bhmm.output_models.gaussian import GaussianOutputModel
from bhmm.util import config
from bhmm.util.statistics import confidence_interval_arr

class GaussianHMM(HMM, GaussianOutputModel):
    r""" Convenience access to an HMM with a Gaussian output model.

    """

    def __init__(self, hmm):
        # superclass constructors
        if not isinstance(hmm.output_model, GaussianOutputModel):
            raise TypeError('Given hmm is not a Gaussian HMM, but has an output model of type: '+
                            str(type(hmm.output_model)))
        GaussianOutputModel.__init__(self, hmm.nstates, means=hmm.output_model.means, sigmas=hmm.output_model.sigmas)
        HMM.__init__(self, hmm.transition_matrix, self, lag=hmm.lag, Pi=hmm.initial_distribution,
                     stationary=hmm.is_stationary, reversible=hmm.is_reversible)


class SampledGaussianHMM(GaussianHMM, SampledHMM):
    """ Sampled Gaussian HMM with a representative single point estimate and error estimates

    Parameters
    ----------
    estimated_hmm : :class:`HMM <generic_hmm.HMM>`
        Representative HMM estimate, e.g. a maximum likelihood estimate or mean HMM.
    sampled_hmms : list of :class:`HMM <generic_hmm.HMM>`
        Sampled HMMs
    conf : float, optional, default = 0.95
        confidence interval, e.g. 0.68 for 1 sigma or 0.95 for 2 sigma.

    """
    def __init__(self, estimated_hmm, sampled_hmms, conf=0.95):
        # enforce right type
        estimated_hmm = GaussianHMM(estimated_hmm)
        for i in range(len(sampled_hmms)):
            sampled_hmms[i] = GaussianHMM(sampled_hmms[i])
        # call GaussianHMM superclass constructer with estimated_hmm
        GaussianHMM.__init__(self, estimated_hmm)
        # call SampledHMM superclass constructor
        SampledHMM.__init__(self, estimated_hmm, sampled_hmms, conf=conf)

    @property
    def means_samples(self):
        r""" Samples of the Gaussian distribution means """
        res = np.empty((self.nsamples, self.nstates, self.dimension), dtype=config.dtype)
        for i in range(self.nsamples):
            for j in range(self.nstates):
                res[i,j,:] = self._sampled_hmms[i].means[j]
        return res

    @property
    def means_mean(self):
        r""" The mean of the Gaussian distribution means """
        return np.mean(self.means_samples, axis=0)

    @property
    def means_std(self):
        r""" The standard deviation of the Gaussian distribution means """
        return np.std(self.means_samples, axis=0)

    @property
    def means_conf(self):
        r""" The standard deviation of the Gaussian distribution means """
        return confidence_interval_arr(self.means_samples, conf=self._conf)

    @property
    def sigmas_samples(self):
        r""" Samples of the Gaussian distribution standard deviations """
        res = np.empty((self.nsamples, self.nstates, self.dimension), dtype=config.dtype)
        for i in range(self.nsamples):
            for j in range(self.nstates):
                res[i,j,:] = self._sampled_hmms[i].sigmas[j]
        return res

    @property
    def sigmas_mean(self):
        r""" The mean of the Gaussian distribution standard deviations """
        return np.mean(self.sigmas_samples, axis=0)

    @property
    def sigmas_std(self):
        r""" The standard deviation of the Gaussian distribution standard deviations """
        return np.std(self.sigmas_samples, axis=0)

    @property
    def sigmas_conf(self):
        r""" The standard deviation of the Gaussian distribution standard deviations """
        return confidence_interval_arr(self.sigmas_samples, conf=self._conf)



__author__ = 'noe'

import numpy as np
import unittest
import bhmm.init.discrete as initdisc


class TestHMM(unittest.TestCase):

    def test_discrete_2_2(self):
        # 2x2 transition matrix
        P = np.array([[0.99,0.01],[0.01,0.99]])
        # generate realization
        import msmtools.generation as msmgen
        T = 10000
        dtrajs = [msmgen.generate_traj(P, T)]
        # estimate initial HMM with 2 states - should be identical to P
        hmm = initdisc.initial_model_discrete(dtrajs, 2)
        # test
        A = hmm.transition_matrix
        B = hmm.output_model.output_probabilities
        # Test stochasticity
        import msmtools.analysis as msmana
        msmana.is_transition_matrix(A)
        np.allclose(B.sum(axis=1), np.ones(B.shape[0]))
        # A should be close to P
        if (B[0,0]<B[1,0]):
            B = B[np.array([1,0]),:]
        assert(np.max(A-P) < 0.01)
        assert(np.max(B-np.eye(2)) < 0.01)

    def test_discrete_4_2(self):
        # 4x4 transition matrix
        nstates = 2
        P = np.array([[0.90, 0.10, 0.00, 0.00],
                      [0.10, 0.89, 0.01, 0.00],
                      [0.00, 0.01, 0.89, 0.10],
                      [0.00, 0.00, 0.10, 0.90]])
        # generate realization
        import msmtools.generation as msmgen
        T = 10000
        dtrajs = [msmgen.generate_traj(P, T)]
        # estimate initial HMM with 2 states - should be identical to P
        hmm = initdisc.initial_model_discrete(dtrajs, nstates)
        # Test if model fit is close to reference. Note that we do not have an exact reference, so we cannot set the
        # tolerance in a rigorous way to test statistical significance. These are just sanity checks.
        Tij = hmm.transition_matrix
        B = hmm.output_model.output_probabilities
        # Test stochasticity
        import msmtools.analysis as msmana
        msmana.is_transition_matrix(Tij)
        np.allclose(B.sum(axis=1), np.ones(B.shape[0]))
        #if (B[0,0]<B[1,0]):
        #    B = B[np.array([1,0]),:]
        Tij_ref = np.array([[0.99, 0.01],
                            [0.01, 0.99]])
        Bref = np.array([[0.5, 0.5, 0.0, 0.0],
                         [0.0, 0.0, 0.5, 0.5]])
        assert(np.max(Tij-Tij_ref) < 0.01)
        assert(np.max(B-Bref) < 0.05 or np.max(B[[1,0]]-Bref) < 0.05)

    def test_discrete_6_3(self):
        # 4x4 transition matrix
        nstates = 3
        P = np.array([[0.90, 0.10, 0.00, 0.00, 0.00, 0.00],
                      [0.20, 0.79, 0.01, 0.00, 0.00, 0.00],
                      [0.00, 0.01, 0.84, 0.15, 0.00, 0.00],
                      [0.00, 0.00, 0.05, 0.94, 0.01, 0.00],
                      [0.00, 0.00, 0.00, 0.02, 0.78, 0.20],
                      [0.00, 0.00, 0.00, 0.00, 0.10, 0.90]])
        # generate realization
        import msmtools.generation as msmgen
        T = 10000
        dtrajs = [msmgen.generate_traj(P, T)]
        # estimate initial HMM with 2 states - should be identical to P
        hmm = initdisc.initial_model_discrete(dtrajs, nstates)
        # Test stochasticity and reversibility
        Tij = hmm.transition_matrix
        B = hmm.output_model.output_probabilities
        import msmtools.analysis as msmana
        msmana.is_transition_matrix(Tij)
        msmana.is_reversible(Tij)
        np.allclose(B.sum(axis=1), np.ones(B.shape[0]))


if __name__=="__main__":
    unittest.main()



import msmtools
__author__ = 'noe'

import warnings

import numpy as np

from bhmm.hmm.generic_hmm import HMM
from bhmm.output_models.discrete import DiscreteOutputModel
from bhmm.util.logger import logger


def initial_model_discrete(observations, nstates, lag=1, reversible=True):
    """Generate an initial model with discrete output densities

    Parameters
    ----------
    observations : list of ndarray((T_i), dtype=int)
        list of arrays of length T_i with observation data
    nstates : int
        The number of states.
    lag : int, optional, default=1
        The lag time to use for initializing the model.

    TODO
    ----
    * Why do we have a `lag` option?  Isn't the HMM model, by definition, lag=1 everywhere?  Why would this be useful instead of just having the user subsample the data?

    Examples
    --------

    Generate initial model for a discrete output model.

    >>> from bhmm import testsystems
    >>> [model, observations, states] = testsystems.generate_synthetic_observations(output_model_type='discrete')
    >>> initial_model = initial_model_discrete(observations, model.nstates)

    """
    # check input
    if not reversible:
        warnings.warn("nonreversible initialization of discrete HMM currently not supported. Using a reversible matrix for initialization.")
        reversible = True

    # estimate Markov model
    C_full = msmtools.estimation.count_matrix(observations, lag)
    lcc = msmtools.estimation.largest_connected_set(C_full)
    Clcc= msmtools.estimation.largest_connected_submatrix(C_full, lcc=lcc)
    T = msmtools.estimation.transition_matrix(Clcc, reversible=True).toarray()

    # pcca
    pcca = msmtools.analysis.dense.pcca.PCCA(T, nstates)

    # default output probability, in order to avoid zero columns
    nstates_full = C_full.shape[0]
    eps = 0.01 * (1.0/nstates_full)

    # Use PCCA distributions, but avoid 100% assignment to any state (prevents convergence)
    B_conn = np.maximum(pcca.output_probabilities, eps)

    # full state space output matrix
    B = eps * np.ones((nstates,nstates_full), dtype=np.float64)
    # expand B_conn to full state space
    B[:,lcc] = B_conn[:,:]
    # renormalize B to make it row-stochastic
    B /= B.sum(axis=1)[:,None]

    # coarse-grained transition matrix
    M = pcca.memberships
    W = np.linalg.inv(np.dot(M.T, M))
    A = np.dot(np.dot(M.T, T), M)
    P_coarse = np.dot(W, A)

    # symmetrize and renormalize to eliminate numerical errors
    X = np.dot(np.diag(pcca.coarse_grained_stationary_probability), P_coarse)
    X = 0.5 * (X + X.T)
    # if there are values < 0, set to eps
    X = np.maximum(X, eps)
    # turn into coarse-grained transition matrix
    A = X / X.sum(axis=1)[:, None]

    logger().info('Initial model: ')
    logger().info('transition matrix = \n'+str(A))
    logger().info('output matrix = \n'+str(B.T))

    # initialize HMM
    # --------------
    output_model = DiscreteOutputModel(B)
    model = HMM(A, output_model)
    return model


